# Copyright 2011 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import logging

from chirp import util, memmap, chirp_common, bitwise, directory, errors
from chirp.drivers.yaesu_clone import YaesuCloneModeRadio

LOG = logging.getLogger(__name__)

CHUNK_SIZE = 16


def _send(s, data):
    for i in range(0, len(data), CHUNK_SIZE):
        chunk = data[i:i+CHUNK_SIZE]
        s.write(chunk)
        echo = s.read(len(chunk))
        if chunk != echo:
            raise Exception("Failed to read echo chunk")


# The IDBLOCK is the first thing sent during an upload or download
# and indicates the radio subtype:
#   USA Unmodified:            b"\x0c\x01\x41\x33\x35\x02\x00\xb8"
#   USA With extended TX mod:  b"\x0c\x01\x41\x33\x35\x03\x00\xb9"
SUPPORTED_IDBLOCKS = [b"\x0c\x01\x41\x33\x35\x02\x00\xb8",
                      b"\x0c\x01\x41\x33\x35\x03\x00\xb9"]
TRAILER = b"\x0c\x02\x41\x33\x35\x00\x00\xb7"
ACK = b"\x0C\x06\x00"


def _download(radio):
    data = b""
    attempts = 30
    for _i in range(0, attempts):
        data = radio.pipe.read(8)
        if data in SUPPORTED_IDBLOCKS:
            radio.subtype = data
            break
        LOG.debug('Download attempt %i received %i: %s',
                  _i, len(data), util.hexprint(data))
        if radio.status_fn:
            status = chirp_common.Status()
            status.max = 1
            status.cur = 0
            status.msg = "Waiting for radio (%i)" % (
                attempts - (_i + 1))
            radio.status_fn(status)

    LOG.debug("Header:\n%s" % util.hexprint(data))

    if len(data) != 8:
        raise Exception("Failed to read header")

    _send(radio.pipe, ACK)

    data = b""

    while len(data) < radio._block_sizes[1]:
        time.sleep(0.1)
        chunk = radio.pipe.read(38)
        LOG.debug("Got: %i:\n%s" % (len(chunk), util.hexprint(chunk)))
        if len(chunk) == 8:
            LOG.debug("END?")
        elif len(chunk) != 38:
            LOG.debug("Should fail?")
            break
            # raise Exception("Failed to get full data block")
        else:
            cs = 0
            for byte in chunk[:-1]:
                cs += byte
            if chunk[-1] != (cs & 0xFF):
                raise Exception("Block failed checksum!")

            data += chunk[5:-1]

        _send(radio.pipe, ACK)
        if radio.status_fn:
            status = chirp_common.Status()
            status.max = radio._block_sizes[1]
            status.cur = len(data)
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    LOG.debug("Total: %i" % len(data))

    return memmap.MemoryMapBytes(data)


def _upload(radio):
    for _i in range(0, 10):
        data = radio.pipe.read(256)
        if not data:
            break
        LOG.debug("What is this garbage?\n%s" % util.hexprint(data))

    _send(radio.pipe, radio.subtype)
    time.sleep(1)
    ack = radio.pipe.read(300)
    LOG.debug("Ack was (%i):\n%s" % (len(ack), util.hexprint(ack)))
    if ack != ACK:
        raise Exception("Radio did not ack ID")

    block = 0
    while block < (radio.get_memsize() // 32):
        data = b"\x0C\x03\x00\x00" + bytes([block])
        data += radio.get_mmap()[block*32:(block+1)*32]
        cs = 0
        for byte in data:
            cs += byte
        data += bytes([cs & 0xFF])

        LOG.debug("Writing block %i:\n%s" % (block, util.hexprint(data)))

        _send(radio.pipe, data)
        time.sleep(0.1)
        ack = radio.pipe.read(3)
        if ack != ACK:
            raise Exception("Radio did not ack block %i" % block)

        if radio.status_fn:
            status = chirp_common.Status()
            status.max = radio._block_sizes[1]
            status.cur = block * 32
            status.msg = "Cloning to radio"
            radio.status_fn(status)
        block += 1

    _send(radio.pipe, TRAILER)


MEM_FORMAT = """
struct {
  bbcd freq[4];
  u8 unknown1[4];
  bbcd offset[2];
  u8 unknown2[2];
  u8 pskip:1,
     skip:1,
     unknown3:1,
     isnarrow:1,
     power:2,
     duplex:2;
  u8 unknown4:6,
     tmode:2;
  u8 tone;
  u8 dtcs;
} memory[200];

#seekto 0x0E00;
struct {
  char name[6];
} names[200];
"""

MODES = ["FM", "NFM"]
TMODES = ["", "Tone", "TSQL", "DTCS"]
DUPLEX = ["", "-", "+", ""]
POWER_LEVELS = [chirp_common.PowerLevel("Hi", watts=65),
                chirp_common.PowerLevel("Mid", watts=25),
                chirp_common.PowerLevel("Low2", watts=10),
                chirp_common.PowerLevel("Low1", watts=5),
                ]
CHARSET = chirp_common.CHARSET_UPPER_NUMERIC + "()+-=*/???|_"


@directory.register
class FT2800Radio(YaesuCloneModeRadio):
    """Yaesu FT-2800"""
    VENDOR = "Yaesu"
    MODEL = "FT-2800M"

    _block_sizes = [8, 7680]
    _memsize = 7680

    @property
    def subtype(self):
        # If our image is from before the subtype was stashed, assume
        # the default unmodified US ID block
        return bytes(self.metadata.get('subtype_idblock',
                                       SUPPORTED_IDBLOCKS[0]))

    @subtype.setter
    def subtype(self, value):
        self.metadata = {'subtype_idblock': [x for x in value]}

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable\n"
            "3. Press and hold in the MHz, Low, and D/MR keys on the radio "
            "while turning it on\n"
            "4. Radio is in clone mode when TX/RX is flashing\n"
            "5. <b>After clicking OK</b>, "
            "press the MHz key on the radio to send"
            " image.\n"
            "    (\"TX\" will appear on the LCD). \n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable\n"
            "3. Press and hold in the MHz, Low, and D/MR keys on the radio "
            "while turning it on\n"
            "4. Radio is in clone mode when TX/RX is flashing\n"
            "5. Press the Low key on the radio "
            "(\"RX\" will appear on the LCD).\n"
            "6. Click OK.")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()

        rf.memory_bounds = (0, 199)

        rf.has_ctone = False
        rf.has_tuning_step = False
        rf.has_dtcs_polarity = False
        rf.has_bank = False

        rf.valid_tuning_steps = [5.0, 10.0, 12.5, 15.0,
                                 20.0, 25.0, 50.0, 100.0]
        rf.valid_modes = MODES
        rf.valid_tmodes = TMODES
        rf.valid_bands = [(137000000, 174000000)]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_duplexes = DUPLEX
        rf.valid_skips = ["", "S", "P"]
        rf.valid_name_length = 6
        rf.valid_characters = CHARSET

        return rf

    def sync_in(self):
        self.pipe.parity = "E"
        self.pipe.timeout = 1
        start = time.time()
        try:
            self._mmap = _download(self)
        except errors.RadioError:
            raise
        except Exception as e:
            LOG.exception('Failed download')
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Downloaded in %.2f sec" % (time.time() - start))
        self.process_mmap()

    def sync_out(self):
        self.pipe.timeout = 1
        self.pipe.parity = "E"
        start = time.time()
        try:
            _upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            LOG.exception('Failed upload')
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Uploaded in %.2f sec" % (time.time() - start))

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _nam = self._memobj.names[number]
        mem = chirp_common.Memory()

        mem.number = number

        if _mem.get_raw(asbytes=False)[0] == "\xFF":
            mem.empty = True
            return mem

        mem.freq = int(_mem.freq) * 10
        mem.offset = int(_mem.offset) * 100000
        mem.duplex = DUPLEX[_mem.duplex]
        mem.tmode = TMODES[_mem.tmode]
        mem.rtone = chirp_common.TONES[_mem.tone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.name = str(_nam.name).rstrip()
        mem.mode = _mem.isnarrow and "NFM" or "FM"
        mem.skip = _mem.pskip and "P" or _mem.skip and "S" or ""
        mem.power = POWER_LEVELS[_mem.power]

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number]
        _nam = self._memobj.names[mem.number]

        if mem.empty:
            _mem.set_raw("\xFF" * (_mem.size() // 8))
            return

        if _mem.get_raw(asbytes=False)[0] == "\xFF":
            # Empty -> Non-empty, so initialize
            _mem.set_raw("\x00" * (_mem.size() // 8))

        # initializing unknowns
        _mem.unknown1 = (0xFF, 0xFF, 0xFF, 0xFF)
        _mem.unknown2 = (0x00, 0x00)
        _mem.unknown3 = 0x01
        _mem.unknown4 = 0x3C

        _mem.freq = mem.freq / 10
        _mem.offset = mem.offset / 100000
        _mem.duplex = DUPLEX.index(mem.duplex)
        _mem.tmode = TMODES.index(mem.tmode)
        _mem.tone = chirp_common.TONES.index(mem.rtone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.isnarrow = MODES.index(mem.mode)
        _mem.pskip = mem.skip == "P"
        _mem.skip = mem.skip == "S"
        if mem.power:
            _mem.power = POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 0

        _nam.name = mem.name.ljust(6)[:6]

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == cls._memsize



# -*- coding: utf-8 -*-
# Copyright 2022 Masen Furer <kf7hvm@0x26.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Lanchonlh HG-UV98 driver written by
#   Masen Furer <kf7hvm@0x26.net>
# With assistance from
#  KG7KMV and Bartłomiej Zieliński
# Based on the implementation of Kenwood TK-8102

import logging
import struct

from chirp import chirp_common, directory, memmap, errors, util
from chirp import bitwise
from chirp.settings import RadioSettingGroup, RadioSetting
from chirp.settings import RadioSettingValueBoolean, RadioSettingValueList
from chirp.settings import RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct {
  lbcd rx_freq[4];
  lbcd tx_freq[4];
  ul16 rx_tone;
  ul16 tx_tone;
  u8 unknown1:6,
     wide:1,
     highpower:1;
  u8 unknown2:5,
     bcl:1,
     scan:1,
     unknown3:1;
  u8 unknown4[2];
} memory[130];

#seekto 0x0a00;
struct {
  u8 abr;  // 0x0a00
  u8 save;
  u8 ch_a_step;
  u8 ch_b_step;
  u8 vox_grd;
  u8 ch_a_sql;
  u8 ch_b_sql;
  u8 roger;
  u8 ch_a_v_m;
  u8 ch_b_v_m;
  u8 ch_a_ch_mdf;
  u8 ch_b_ch_mdf;
  u8 tdr;
  u8 unknown5[3];
  u8 unknown6[5];  // 0x0a10
  u8 english;
  u8 beep;
  u8 voice;
  u8 night_mode;
  u8 abr_lv;  // backlight level
  u8 tot;
  u8 toa;
  u8 vox_dly;
  u8 sc_rev;
  u8 lockmode;
  u8 autolock;
  u8 unknown7;  // 0x0a20
  u8 pf1_short;
  u8 pf1_long;
  u8 pf2_short;
  u8 pf2_long;
  u8 top_short;
  u8 top_long;
  u8 rpt_rct;
  u8 sc_qt;
  u8 pri_ch;
  u8 pri_scn;
  u8 unknown8;
  u8 aprs_rx_band;
  u8 ch_a_mute;
  u8 ch_b_mute;
  u8 unknown9[7];  // 0x0a30
  u8 tx_priority;
  u8 aprs_rx_popup;
  u8 aprs_rx_tone;
  u8 aprs_tx_tone;
  u8 unknown10;
  u8 auto_lock_dly;
  u8 menu_dly;
  u8 beacon_exit_dly;
  u8 unknown11;
  u8 unknown12[2];        // 0x0a40
  u8 ch_a_mem_ch;
  u8 ch_b_mem_ch;
  u8 unknown13[12];
} settings;

#seekto 0x1000;
struct {
  char name[11];
  u8 unknown[5];
} name[128];

struct {
    char callsign[9];
    u8 null;
} aprs;

#seekto 0x1f80;
struct {
  ul32 unknown[8];
} unknown_settings;

"""

BOUNDS = [(136000000, 174000000), (400000000, 500000000)]
OFFSETS = [600000, 5000000]
MAX_CHANNELS = 128
MAX_NAME = 8
NAME_FIELD_SIZE = 11
CHUNK_SIZE = 64
MAX_ADDR = 0x2000
POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1),
                chirp_common.PowerLevel("High", watts=5)]
MODES = ["FM", "NFM"]
SPECIAL_CHANNELS = ['VFO-A', 'VFO-B']

# Settings maps
ABR_LIST = [str(v) for v in range(0, 151, 5)]
STEP_LIST = ["5.0", "6.25", "10.0", "12.5", "25.0", "50.0", "100.0"]
VOX_LIST = ["OFF", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
SQL_LIST = [str(v) for v in range(0, 10)]
ROGER_LIST = ["OFF", "BEGIN", "END", "BOTH"]
MDF_LIST = ["NUM+FREQ", "NUMBER", "NAME"]
VM_LIST = ["VFO", "MEMORY"]
LANG_LIST = ["CHINESE", "ENGLISH"]
ABR_LV_LIST = [str(v) for v in range(1, 11)]
TOT_LIST = [str(v) for v in range(0, 601, 15)]
TOA_LIST = ["OFF"] + ["{} S".format(v) for v in range(1, 11)]
VOX_DLY_LIST = [str(v) for v in range(1, 11)]
SC_REV_LIST = ["TIME", "BUSY", "HOLD"]
LOCKMODE_LIST = ["KEY", "KEY+DIAL", "KEY+DIAL+PTT"]
AUTOLOCK_LIST = ["AUTO", "Manual"]
PF1_LIST = ["BACK LIGHT", "SCAN", "SQUELCH", "TORCH", "BAND A/B"]
PF2_LIST = ["BEACON", "LIST", "TORCH", "BACK LIGHT"]
TOP_LIST = ["ALERT", "REMOTE ALERT", "TORCH", "TXP", "CH MDF", "OFF_NET_KEY"]
SC_QT_LIST = ["Decode", "Encode", "Decode+Encode"]
APRS_RX_LIST = ["OFF", "BAND A", "BAND B"]
TX_PRIORITY_LIST = ["VOICE", "APRS"]
AUTOLOCK_DLY_LIST = ["{} S".format(v) for v in range(5, 31)]
BEACON_EXIT_DLY_LIST = MENU_DLY_LIST = AUTOLOCK_DLY_LIST


def make_frame(cmd, addr, length, data=b""):
    if not isinstance(data, bytes):
        data = data.decode("ascii")
    return struct.pack(">BHB", ord(cmd), addr, length) + data


def send(radio, frame):
    LOG.debug("%04i P>R: %s" % (len(frame), util.hexprint(frame)))
    radio.pipe.write(frame)


def recv(radio, readdata=True):
    hdr = radio.pipe.read(4)
    cmd, addr, length = struct.unpack(">BHB", hdr)
    if readdata:
        data = radio.pipe.read(length)
        LOG.debug("     P<R: %s" % util.hexprint(hdr + data))
        if len(data) != length:
            raise errors.RadioError("Radio sent %i bytes (expected %i)" % (
                    len(data), length))
    else:
        data = b""
    radio.pipe.write(b"\x06")
    ack = radio.pipe.read(1)
    if ack != b"\x06":
        raise errors.RadioError("Radio didn't ack our read ack")
    return addr, bytes(data)


def do_ident(radio):
    send(radio, b"NiNHSG0N")
    ack = radio.pipe.read(1)
    if ack != b"\x06":
        raise errors.RadioError("Radio refused program mode: {}".format(ack))
    radio.pipe.write(b"\x02")
    ident = radio.pipe.read(8)
    LOG.debug('ident string was %r' % ident)
    if ident != radio.IDENT:
        raise errors.RadioError(
            "Incorrect model: %s, expected %r" % (
                util.hexprint(ident), radio.IDENT))
    LOG.info("Model: %s (%s)" % (radio.MODEL, util.hexprint(ident)))
    radio.pipe.write(b"\x06")
    ack = radio.pipe.read(1)
    if ack != b"\x06":
        raise errors.RadioError("Radio entered program mode, but didn't ack our ack")


def do_download(radio):
    radio.pipe.parity = "E"
    radio.pipe.timeout = 1
    do_ident(radio)

    data = bytes(b"")
    for addr in range(0, MAX_ADDR, CHUNK_SIZE):
        send(radio, make_frame(bytes(b"R"), addr, CHUNK_SIZE))
        _addr, _data = recv(radio)
        if _addr != addr:
            raise errors.RadioError("Radio sent unexpected address")
        data += _data
        radio.pipe.write(b"\x06")
        ack = radio.pipe.read(1)
        if ack != b"\x06":
            raise errors.RadioError("Radio refused block at %04x" % addr)

        status = chirp_common.Status()
        status.cur = addr
        status.max = MAX_ADDR
        status.msg = "Cloning from radio"
        radio.status_fn(status)

    radio.pipe.write(b"\x45")
    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    radio.pipe.parity = "E"
    radio.pipe.timeout = 1
    do_ident(radio)

    mmap = radio._mmap
    for addr in range(0, MAX_ADDR, CHUNK_SIZE):
        send(radio, make_frame(b"W", addr, CHUNK_SIZE, mmap[addr:addr + CHUNK_SIZE]))
        ack = radio.pipe.read(1)
        if ack != b"\x06":
            raise errors.RadioError("Radio refused block at %04x" % addr)
        radio.pipe.write(b"\x06")
        ack = radio.pipe.read(1)
        if ack != b"\x06":
            raise errors.RadioError("Radio didn't ack our read ack")

        status = chirp_common.Status()
        status.cur = addr
        status.max = MAX_ADDR
        status.msg = "Cloning to radio"
        radio.status_fn(status)

    radio.pipe.write(b"\x45")


def offset_for(freq):
    for bounds, offset in zip(BOUNDS, OFFSETS):
        if bounds[0] <= freq <= bounds[1]:
            return offset
    return 0


class RadioSettingValueChannel(RadioSettingValueList):
    """A setting that points to a defined channel."""
    def __init__(self, radio, current_raw):
        current = int(current_raw)
        current_mem = radio.get_memory(current)
        lo, hi = radio.get_features().memory_bounds
        options = [
            self._format_memory(mem)
            for mem in [radio.get_memory(n)
                        for n in range(lo, hi + 1)]]
        RadioSettingValueList.__init__(self, options,
                                       self._format_memory(current_mem))

    @staticmethod
    def _format_memory(m):
        if m.empty:
            return str(int(m.number))
        return "%i %.4f %s" % (m.number, m.freq / 1e6, m.name)

    def __int__(self):
        return int(self.get_value().partition(" ")[0])


@directory.register
class LanchonlhHG_UV98(chirp_common.CloneModeRadio, chirp_common.ExperimentalRadio):
    """
    Lanchonlh HG-UV98

    Memory map decoding by KG7KMV
    Chirp integration by KF7HVM
    """
    VENDOR = "Lanchonlh"
    MODEL = "HG-UV98"
    IDENT = b"P3107\0\0\0"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    _upper = MAX_CHANNELS

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = (
            "This Lanchonlh HG-UV98 driver is an alpha version. "
            "Proceed with Caution and backup your data. "
            "Always confirm the correctness of your settings with the "
            "official programming tool.")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_cross = False
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.has_name = True
        rf.has_rx_dtcs = True
        rf.valid_characters = chirp_common.CHARSET_ALPHANUMERIC
        rf.valid_tuning_steps = [5.0, 6.25, 10.0, 12.5, 20.0, 25.0, 50.0, 100.0]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_modes = MODES
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_skips = ["", "S"]
        rf.valid_bands = [(136000000, 174000000), (400000000, 500000000)]
        rf.valid_name_length = 8
        rf.valid_special_chans = SPECIAL_CHANNELS
        rf.memory_bounds = (1, self._upper)
        return rf

    def sync_in(self):
        try:
            self._mmap = do_download(self)
        except errors.RadioError:
            self.pipe.write(b"\x45")
            raise
        except Exception as e:
            raise errors.RadioError("Failed to download from radio: %s" % e)
        self.process_mmap()

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_out(self):
        try:
            do_upload(self)
        except errors.RadioError:
            self.pipe.write(b"\x45")
            raise
        except Exception as e:
            raise errors.RadioError("Failed to upload to radio: %s" % e)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _get_tone(self, _mem, mem):
        def _get_dcs(val):
            code = int("%03o" % (val & 0x07FF))
            pol = (val & 0x8000) and "R" or "N"
            return code, pol

        if _mem.tx_tone != 0xFFFF and _mem.tx_tone > 0x2800:
            tcode, tpol = _get_dcs(_mem.tx_tone)
            mem.dtcs = tcode
            txmode = "DTCS"
        elif _mem.tx_tone != 0xFFFF:
            mem.rtone = _mem.tx_tone / 10.0
            txmode = "Tone"
        else:
            txmode = ""

        if _mem.rx_tone != 0xFFFF and _mem.rx_tone > 0x2800:
            rcode, rpol = _get_dcs(_mem.rx_tone)
            mem.rx_dtcs = rcode
            rxmode = "DTCS"
        elif _mem.rx_tone != 0xFFFF:
            mem.ctone = _mem.rx_tone / 10.0
            rxmode = "Tone"
        else:
            rxmode = ""

        if txmode == "Tone" and not rxmode:
            mem.tmode = "Tone"
        elif txmode == rxmode and txmode == "Tone" and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif txmode == rxmode and txmode == "DTCS" and mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif rxmode or txmode:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        if mem.tmode == "DTCS":
            mem.dtcs_polarity = "%s%s" % (tpol, rpol)

    def get_memory(self, number):

        mem = chirp_common.Memory()

        if isinstance(number, str):
            mem.number = MAX_CHANNELS + SPECIAL_CHANNELS.index(number) + 1
            mem.extd_number = number
        elif number > MAX_CHANNELS:
            mem.number = number
        else:
            mem.number = number
            _name = self._memobj.name[mem.number - 1]

        _mem = self._memobj.memory[mem.number - 1]

        if mem.number > MAX_CHANNELS:
            mem.immutable = ['name']
        else:
            mem.name, _, _ = _name.name.get_raw().partition(b"\xFF")
            mem.name = mem.name.decode('ascii').rstrip()

        if _mem.get_raw()[:4] == b"\xFF\xFF\xFF\xFF":
            mem.empty = True
            return mem

        mem.freq = int(_mem.rx_freq) * 10
        offset = (int(_mem.tx_freq) * 10) - mem.freq
        if offset < 0:
            mem.offset = abs(offset)
            mem.duplex = "-"
        elif offset > 0:
            mem.offset = offset
            mem.duplex = "+"
        else:
            mem.offset = 0

        self._get_tone(_mem, mem)
        mem.power = POWER_LEVELS[_mem.highpower]
        mem.mode = MODES[_mem.wide]
        mem.skip = not _mem.scan and "S" or ""

        mem.extra = RadioSettingGroup("all", "All Settings")

        bcl = RadioSetting("bcl", "Busy Channel Lockout",
                           RadioSettingValueBoolean(bool(_mem.bcl)))
        mem.extra.append(bcl)

        return mem

    def _set_tone(self, mem, _mem):
        def _set_dcs(code, pol):
            val = int("%i" % code, 8) + 0x2800
            if pol == "R":
                val += 0xA000
            return val

        rx_mode = tx_mode = None
        rx_tone = tx_tone = 0xFFFF

        if mem.tmode == "Tone":
            tx_mode = "Tone"
            rx_mode = None
            tx_tone = int(mem.rtone * 10)
        elif mem.tmode == "TSQL":
            rx_mode = tx_mode = "Tone"
            rx_tone = tx_tone = int(mem.ctone * 10)
        elif mem.tmode == "DTCS":
            tx_mode = rx_mode = "DTCS"
            tx_tone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            rx_tone = _set_dcs(mem.dtcs, mem.dtcs_polarity[1])
        elif mem.tmode == "Cross":
            tx_mode, rx_mode = mem.cross_mode.split("->")
            if tx_mode == "DTCS":
                tx_tone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            elif tx_mode == "Tone":
                tx_tone = int(mem.rtone * 10)
            if rx_mode == "DTCS":
                rx_tone = _set_dcs(mem.rx_dtcs, mem.dtcs_polarity[1])
            elif rx_mode == "Tone":
                rx_tone = int(mem.ctone * 10)

        _mem.rx_tone = rx_tone
        _mem.tx_tone = tx_tone

        LOG.debug("Set TX %s (%i) RX %s (%i)" %
                  (tx_mode, _mem.tx_tone, rx_mode, _mem.rx_tone))

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]

        if mem.empty:
            _mem.set_raw(b"\xFF" * 16)
            return

        if mem.number < 129:
            _name = self._memobj.name[mem.number - 1]
            _namelength = self.get_features().valid_name_length
            for i in range(NAME_FIELD_SIZE):
                try:
                    _name.name[i] = mem.name[i]
                except IndexError:
                    _name.name[i] = "\xFF"

        # clear reserved fields
        _mem.unknown1 = 0xFF
        _mem.unknown2 = 0xFF
        _mem.unknown3 = 0xFF
        _mem.unknown4 = (0xFF, 0xFF)
        _mem.rx_freq = mem.freq / 10
        mem_offset = mem.offset or offset_for(mem.freq)
        if mem.duplex == "+":
            _mem.tx_freq = (mem.freq + mem_offset) / 10
        elif mem.duplex == "-":
            _mem.tx_freq = (mem.freq - mem_offset) / 10
        else:
            _mem.tx_freq = mem.freq / 10

        self._set_tone(mem, _mem)

        _mem.highpower = mem.power == POWER_LEVELS[1]
        _mem.wide = mem.mode == "NFM"
        _mem.scan = mem.skip != "S"

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _mem = self._memobj
        _settings = _mem.settings
        basic = RadioSettingGroup("basic", "Basic")
        display = RadioSettingGroup("display", "Display")
        scan = RadioSettingGroup("scan", "Scan")
        buttons = RadioSettingGroup("buttons", "Buttons")
        vfo = RadioSettingGroup("vfo", "VFO")
        advanced = RadioSettingGroup("advanced", "Advanced")
        aprs = RadioSettingGroup("aprs", "APRS")
        top = RadioSettings(basic, display, scan, buttons, vfo, advanced, aprs)

        basic.append(
            RadioSetting("save", "Power Save",
                         RadioSettingValueBoolean(_settings.save)))
        basic.append(
            RadioSetting("roger", "Roger Beep",
                         RadioSettingValueList(ROGER_LIST,
                                               ROGER_LIST[_settings.roger])))
        basic.append(
            RadioSetting("beep", "System Beep",
                         RadioSettingValueBoolean(_settings.beep)))
        basic.append(
            RadioSetting("tot", "Timeout Timer (sec)",
                         RadioSettingValueList(TOT_LIST,
                                               TOT_LIST[_settings.tot])))
        basic.append(
            RadioSetting("toa", "Timeout Timer Alarm",
                         RadioSettingValueList(TOA_LIST,
                                               TOA_LIST[_settings.toa])))
        basic.append(
            RadioSetting("lockmode", "Lock Mode",
                         RadioSettingValueList(
                            LOCKMODE_LIST,
                            LOCKMODE_LIST[_settings.lockmode])))
        basic.append(
            RadioSetting("autolock", "Auto Lock",
                         RadioSettingValueList(
                            AUTOLOCK_LIST,
                            AUTOLOCK_LIST[_settings.autolock])))
        basic.append(
            RadioSetting("auto_lock_dly", "Auto Lock Delay",
                         RadioSettingValueList(
                            AUTOLOCK_DLY_LIST,
                            AUTOLOCK_DLY_LIST[_settings.auto_lock_dly])))
        display.append(
            RadioSetting("abr", "Screen Save",
                         RadioSettingValueList(ABR_LIST,
                                               ABR_LIST[_settings.abr])))
        display.append(
            RadioSetting("abr_lv", "Back Light Brightness",
                         RadioSettingValueList(ABR_LV_LIST,
                                               ABR_LV_LIST[_settings.abr_lv])))
        display.append(
            RadioSetting("night_mode", "Night Mode (Light on Dark)",
                         RadioSettingValueBoolean(_settings.night_mode)))
        display.append(
            RadioSetting("menu_dly", "Menu Delay",
                         RadioSettingValueList(
                            MENU_DLY_LIST,
                            MENU_DLY_LIST[_settings.menu_dly])))
        display.append(
            RadioSetting("english", "Language",
                         RadioSettingValueList(LANG_LIST,
                                               LANG_LIST[_settings.english])))
        scan.append(
            RadioSetting("pri_scn", "Priority Scan",
                         RadioSettingValueBoolean(_settings.pri_scn)))
        scan.append(
            RadioSetting("pri_ch", "Priority Channel",
                         RadioSettingValueChannel(self, _settings.pri_ch)))
        scan.append(
            RadioSetting("sc_rev", "Scan Resume",
                         RadioSettingValueList(SC_REV_LIST,
                                               SC_REV_LIST[_settings.sc_rev])))
        scan.append(
            RadioSetting("sc_qt", "Code Save",
                         RadioSettingValueList(SC_QT_LIST,
                                               SC_QT_LIST[_settings.sc_qt])))
        buttons.append(
            RadioSetting("pf1_short", "PF1 (Side, Upper) Button Short Press",
                         RadioSettingValueList(PF1_LIST,
                                               PF1_LIST[_settings.pf1_short])))
        buttons.append(
            RadioSetting("pf1_long", "PF1 (Side, Upper) Button Long Press",
                         RadioSettingValueList(PF1_LIST,
                                               PF1_LIST[_settings.pf1_long])))
        buttons.append(
            RadioSetting("pf2_short", "PF2 (Side, Lower) Button Short Press",
                         RadioSettingValueList(PF2_LIST,
                                               PF2_LIST[_settings.pf2_short])))
        buttons.append(
            RadioSetting("pf2_long", "PF2 (Side, Lower) Button Long Press",
                         RadioSettingValueList(PF2_LIST,
                                               PF2_LIST[_settings.pf2_long])))
        buttons.append(
            RadioSetting("top_short", "Top Button Short Press",
                         RadioSettingValueList(TOP_LIST,
                                               TOP_LIST[_settings.top_short])))
        buttons.append(
            RadioSetting("top_long", "Top Button Long Press",
                         RadioSettingValueList(TOP_LIST,
                                               TOP_LIST[_settings.top_long])))
        vfo.append(
            RadioSetting("tdr", "VFO B Enabled",
                         RadioSettingValueBoolean(_settings.tdr)))
        vfo.append(
            RadioSetting("ch_a_step", "VFO Frequency Step (A)",
                         RadioSettingValueList(
                            STEP_LIST,
                            STEP_LIST[_settings.ch_a_step])))
        vfo.append(
            RadioSetting("ch_b_step", "VFO Frequency Step (B)",
                         RadioSettingValueList(
                            STEP_LIST,
                            STEP_LIST[_settings.ch_b_step])))
        vfo.append(
            RadioSetting("ch_a_sql", "Squelch (A)",
                RadioSettingValueList(SQL_LIST,
                                      SQL_LIST[_settings.ch_a_sql])))
        vfo.append(
            RadioSetting("ch_b_sql", "Squelch (B)",
                         RadioSettingValueList(SQL_LIST,
                                               SQL_LIST[_settings.ch_b_sql])))
        vfo.append(
            RadioSetting("ch_a_mem_ch", "Memory Channel (A)",
                         RadioSettingValueChannel(self,
                                                  _settings.ch_a_mem_ch)))
        vfo.append(
            RadioSetting("ch_b_mem_ch", "Memory Channel (B)",
                         RadioSettingValueChannel(self,
                                                  _settings.ch_b_mem_ch)))
        vfo.append(
            RadioSetting("ch_a_ch_mdf", "Memory Display Format (A)",
                         RadioSettingValueList(
                            MDF_LIST,
                            MDF_LIST[_settings.ch_a_ch_mdf])))
        vfo.append(
            RadioSetting("ch_b_ch_mdf", "Memory Display Format (B)",
                         RadioSettingValueList(
                            MDF_LIST,
                            MDF_LIST[_settings.ch_b_ch_mdf])))
        vfo.append(
            RadioSetting("ch_a_v_m", "VFO/MEM (A)",
                         RadioSettingValueList(
                             VM_LIST, VM_LIST[_settings.ch_a_v_m])))
        vfo.append(
            RadioSetting("ch_b_v_m", "VFO/MEM (B)",
                         RadioSettingValueList(
                             VM_LIST, VM_LIST[_settings.ch_b_v_m])))
        advanced.append(
            RadioSetting("vox_grd", "VOX Sensitivity",
                         RadioSettingValueList(
                             VOX_LIST, VOX_LIST[_settings.vox_grd])))
        advanced.append(
            RadioSetting("vox_dly", "VOX Delay",
                         RadioSettingValueList(
                             VOX_DLY_LIST, VOX_DLY_LIST[_settings.vox_dly])))
        advanced.append(
            RadioSetting("voice", "Voice Assist",
                         RadioSettingValueBoolean(_settings.voice)))
        advanced.append(
            RadioSetting("rpt_rct", "RPT Roger",
                         RadioSettingValueBoolean(_settings.rpt_rct)))
        aprs.append(
            RadioSetting("aprs_rx_band", "RX Band",
                         RadioSettingValueList(
                             APRS_RX_LIST,
                             APRS_RX_LIST[_settings.aprs_rx_band])))
        aprs.append(
            RadioSetting("ch_a_mute", "Band A Mute",
                         RadioSettingValueBoolean(_settings.ch_a_mute)))
        aprs.append(
            RadioSetting("ch_b_mute", "Band B Mute",
                         RadioSettingValueBoolean(_settings.ch_b_mute)))
        aprs.append(
            RadioSetting("tx_priority", "TX Priority",
                RadioSettingValueList(
                    TX_PRIORITY_LIST,
                    TX_PRIORITY_LIST[_settings.tx_priority])))
        aprs.append(
            RadioSetting("aprs_rx_popup", "APRS Popup",
                         RadioSettingValueBoolean(_settings.aprs_rx_popup)))
        aprs.append(
            RadioSetting("aprs_rx_tone", "RX Tone",
                         RadioSettingValueBoolean(_settings.aprs_rx_tone)))
        aprs.append(
            RadioSetting("aprs_tx_tone", "TX Tone",
                         RadioSettingValueBoolean(_settings.aprs_tx_tone)))
        aprs.append(
            RadioSetting("beacon_exit_dly", "Beacon Message Delay",
                         RadioSettingValueList(
                             BEACON_EXIT_DLY_LIST,
                             BEACON_EXIT_DLY_LIST[_settings.beacon_exit_dly])))

        return top

    def set_settings(self, settings):
        _mem = self._memobj
        _settings = _mem.settings

        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            name = element.get_name()
            value = element.value

            if hasattr(_settings, name):
                setattr(_settings, name, value)



# Copyright 2012 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import time
import logging


from chirp import chirp_common, errors, util, directory, memmap
from chirp import bitwise
from chirp.settings import InvalidValueError, RadioSetting, \
    RadioSettingGroup, RadioSettingValueFloat, \
    RadioSettingValueList, RadioSettingValueBoolean, \
    RadioSettingValueString, RadioSettings
from textwrap import dedent
from chirp import bandplan_na

LOG = logging.getLogger(__name__)


MEM_FORMAT = """
#seekto 0x0008;
struct {
  lbcd rxfreq[4];
  lbcd txfreq[4];
  lbcd rxtone[2];
  lbcd txtone[2];
  u8 unused1;
  u8 pttid:2,
     freqhop:1,
     unused3:1,
     unused4:1,
     bcl:1,
     unused5:1,
     unused2:1;
  u8 unused6:1,
     unused7:1,
     lowpower:2,
     wide:1,
     unused8:1,
     offset:2;
  u8 unused10;
} memory[200];

#seekto 0x0D38;
struct {
  char name[8];
  u8 unknown2[8];
} names[200];

#seekto 0x0CA8;
struct {
  u8 txled:1,
     rxled:1,
     unused11:1,
     ham:1,
     gmrs:1,
     unused14:1,
     unused15:1,
     pritx:1;
  u8 scanmode:2,
     unused16:1,
     keyautolock:1,
     unused17:1,
     btnvoice:1,
     unknown18:1,
     voiceprompt:1;
  u8 fmworkmode:1,
     sync:1,
     tonevoice:2,
     fmrec:1,
     mdfa:1,
     aworkmode:2;
  u8 ponmsg:2,
     unused19:1,
     mdfb:1,
     unused20:1,
     dbrx:1,
     bworkmode:2;
  u8 ablock;
  u8 bblock;
  u8 fmroad;
  u8 unused21:1,
     tailclean:1,
     rogerprompt:1,
     unused23:1,
     unused24:1,
     voxgain:3;
  u8 astep:4,
     bstep:4;
  u8 squelch;
  u8 tot;
  u8 lang;
  u8 save;
  u8 ligcon;
  u8 voxdelay;
  u8 onlychmode:1,
     unused:6,
     alarm:1;
} settings;

#seekto 0x0CB8;
struct {
    u8 ofseta[4];
    u8 ununsed26[12];
} aoffset;

#seekto 0x0CBC;
struct {
    u8 ofsetb[4];
    u8 ununsed27[12];
} boffset;

#seekto 0x0CD8;
struct{
    u8 fmblock[4];
}fmmode[25];

#seekto 0x1A08;
struct{
    u8 block8:1,
       block7:1,
       block6:1,
       block5:1,
       block4:1,
       block3:1,
       block2:1,
       block1:1;
} usedflags[25];

#seekto 0x1a28;
struct{
  u8 scan8:1,
     scan7:1,
     scan6:1,
     scan5:1,
     scan4:1,
     scan3:1,
     scan2:1,
     scan1:1;
} scanadd[25];

#seekto 0x1B38;
struct{
    u8 vfo[4];
}fmvfo;

#seekto 0x1B58;
struct {
  lbcd rxfreqa[4];
  lbcd txfreq[4];
  u8 rxtone[2];
  u8 txtone[2];
  u8 unused1;
  u8 pttid:2,
     specialqta:1,
     unused3:1,
     unused4:1,
     bcl:1,
     unused5:1,
     unused2:1;
  u8 unused6:1,
     unused7:1,
     lowpower:2,
     wide:1,
     unused8:1,
     offset:2;
  u8 unused10;
} vfoa;

#seekto 0x1B68;
struct {
  lbcd rxfreqb[4];
  lbcd txfreq[4];
  u8 rxtoneb[2];
  u8 txtone[2];
  u8 unused1;
  u8 pttid:2,
     specialqtb:1,
     unused3:1,
     unused4:1,
     bclb:1,
     unused5:1,
     unused2:1;
  u8 unused6:1,
     unused7:1,
     lowpowerb:2,
     wideb:1,
     unused8:1,
     offsetb:2;
  u8 unused10;
} vfob;

#seekto 0x1B78;
struct{
   u8  block8:1,
       block7:1,
       block6:1,
       block5:1,
       block4:1,
       block3:1,
       block2:1,
       block1:1;
} fmusedflags[4];

#seekto 0x1c08;
struct {
  char msg1[16];
  char msg2[16];
  char msg3[16];
} poweron_msg;


#seekto 0x1CC8;
struct{
  u8 stopkey1;
  u8 ssidekey1;
  u8 ssidekey2;
  u8 ltopkey2;
  u8 lsidekey3;
  u8 lsidekey4;
  u8 unused25[10];
} press;

#seekto 0x1E28;
struct{
    u8 idcode[3];
}icode;

#seekto 0x1E31;
struct{
    u8 gcode;
}groupcode;

#seekto 0x1E38;
struct{
    u8 group1[7];
}group1;

#seekto 0x1E48;
struct{
    u8 group2[7];
}group2;

#seekto 0x1E58;
struct{
    u8 group3[7];
}group3;

#seekto 0x1E68;
struct{
    u8 group4[7];
}group4;

#seekto 0x1E78;
struct{
    u8 group5[7];
}group5;

#seekto 0x1E88;
struct{
    u8 group6[7];
}group6;

#seekto 0x1E98;
struct{
    u8 group7[7];
}group7;

#seekto 0x1EA8;
struct{
    u8 group8[7];
}group8;

#seekto 0x1EC8;
struct{
    u8 scode[7];
}startcode;

#seekto 0x1ED8;
struct{
    u8 ecode[7];
}endcode;
"""

MEM_FORMAT_H3 = """
#seekto 0x0008;
struct {
  lbcd rxfreq[4];
  lbcd txfreq[4];
  lbcd rxtone[2];
  lbcd txtone[2];
  u8 unused1;
  u8 pttid:2,
     freqhop:1,
     unused3:1,
     unused4:1,
     bcl:1,
     unused5:1,
     unused2:1;
  u8 unused6:1,
     scramble:1,
     lowpower:2,
     wide:1,
     unused8:1,
     offset:2;
  u8 unused10;
} memory[200];

#seekto 0x0C98;
struct{
  u8 stopkey1;
  u8 ssidekey1;
  u8 ssidekey2;
  u8 ltopkey2;
  u8 lsidekey3;
  u8 lsidekey4;
} press;

#seekto 0x0CA8;
struct {
  u8 txled:1,
     rxled:1,
     unused11:1,
     ham:1,
     gmrs:1,
     unused14:1,
     unused15:1,
     pritx:1;
  u8 scanmode:2,
     unused16:1,
     keyautolock:1,
     unused17:1,
     btnvoice:1,
     unknown18:1,
     voiceprompt:1;
  u8 fmworkmode:1,
     sync:1,
     tonevoice:2,
     fmrec:1,
     mdfa:1,
     aworkmode:2;
  u8 ponmsg:2,
     unused19:1,
     mdfb:1,
     unused20:1,
     dbrx:1,
     bworkmode:2;
  u8 ablock;
  u8 bblock;
  u8 fmroad;
  u8 unused21:1,
     tailclean:1,
     rogerprompt_:1,
     kill:1,
     stun:1,
     voxgain:3;
  u8 astep:4,
     bstep:4;
  u8 squelch;
  u8 tot;
  u8 rogerprompt:2,
     unused11_4:4,
     lang:1,
     unused11_1:1;
  u8 save;
  u8 ligcon;
  u8 voxdelay;
  u8 onlychmode:1,
     breathled:3,
     unused:3,
     alarm:1;
} settings;

#seekto 0x0CB8;
struct {
    u8 ofseta[4];
    u8 ununsed26[12];
} aoffset;

#seekto 0x0CBC;
struct {
    u8 ofsetb[4];
    u8 ununsed27[12];
} boffset;

#seekto 0x0CD8;
struct{
    u8 fmblock[4];
}fmmode[25];


#seekto 0x0D48;
struct {
  char name[8];
} names[200];

#seekto 0x1808;
struct{
    u8 stuncode[16];
    u8 killcode[16];
}skcode;

#seekto 0x1828;
struct{
    u8 idcode[3];
}icode;

#seekto 0x1837;
struct{
    u8 gcode;
}groupcode;

#seekto 0x1838;
struct{
    u8 group1[7];
}group1;

#seekto 0x1848;
struct{
    u8 group2[7];
}group2;

#seekto 0x1858;
struct{
    u8 group3[7];
}group3;

#seekto 0x1868;
struct{
    u8 group4[7];
}group4;

#seekto 0x1878;
struct{
    u8 group5[7];
}group5;

#seekto 0x1888;
struct{
    u8 group6[7];
}group6;

#seekto 0x1898;
struct{
    u8 group7[7];
}group7;

#seekto 0x18A8;
struct{
    u8 group8[7];
}group8;

#seekto 0x18C8;
struct{
    u8 scode[7];
}startcode;

#seekto 0x18D8;
struct{
    u8 ecode[7];
}endcode;

#seekto 0x1908;
struct{
    u8 block8:1,
       block7:1,
       block6:1,
       block5:1,
       block4:1,
       block3:1,
       block2:1,
       block1:1;
} usedflags[25];

#seekto 0x1928;
struct{
  u8 scan8:1,
     scan7:1,
     scan6:1,
     scan5:1,
     scan4:1,
     scan3:1,
     scan2:1,
     scan1:1;
} scanadd[25];

#seekto 0x1948;
struct{
  u8  block8:1,
       block7:1,
       block6:1,
       block5:1,
       block4:1,
       block3:1,
       block2:1,
       block1:1;
} fmusedflags[4];

#seekto 0x1958;
struct {
  lbcd rxfreqa[4];
  lbcd txfreq[4];
  u8 rxtone[2];
  u8 txtone[2];
  u8 unused1;
  u8 pttid:2,
     specialqta:1,
     unused3:1,
     unused4:1,
     bcl:1,
     unused5:1,
     unused2:1;
  u8 unused6:1,
     unused7:1,
     lowpower:2,
     wide:1,
     unused8:1,
     offset:2;
  u8 unused10;
} vfoa;

#seekto 0x1968;
struct {
  lbcd rxfreqb[4];
  lbcd txfreq[4];
  u8 rxtoneb[2];
  u8 txtone[2];
  u8 unused1;
  u8 pttid:2,
     specialqtb:1,
     unused3:1,
     unused4:1,
     bclb:1,
     unused5:1,
     unused2:1;
  u8 unused6:1,
     unused7:1,
     lowpowerb:2,
     wideb:1,
     unused8:1,
     offsetb:2;
  u8 unused10;
} vfob;

#seekto 0x1978;
struct{
    u8 vfo[4];
}fmvfo;


#seekto 0x1c08;
struct {
  char msg1[16];
  char msg2[16];
  char msg3[16];
} poweron_msg;

#seekto 0x1f28;
struct{
  u8 micgain;
} mic;

"""

# basic settings
SQUELCH = ['%s' % x for x in range(0, 10)]
LIGHT_LIST = ["CONT", "5s", "10s", "15s", "30s"]
VOICE_PRMPT_LIST = ["OFF", "ON"]
AUTOLOCK_LIST = ["OFF", "ON"]
TIME_OUT_LIST = ["OFF", "60s", "120s", "180s"]
MDFA_LIST = ["Frequency", "Name"]
MDFB_LIST = ["Frequency", "Name"]
SYNC_LIST = ["ON", "OFF"]
LANG_LIST = ["Chinese", "English"]
BTV_SAVER_LIST = ["OFF", "1:1", "1:2", "1:3", "1:4"]
DBRX_LIST = ["OFF", "ON"]
ASTEP_LIST = ["2.50K", "5.00K", "6.25K",
              "10.00K", "12.00K", "25.00K", "50.00K"]
BSTEP_LIST = ["2.50K", "5.00K", "6.25K",
              "10.00K", "12.00K", "25.00K", "50.00K"]
SCAN_MODE_LIST = ["TO", "CO", "SE"]
PRIO_LIST = ["Edit", "Busy"]
SHORT_KEY_LIST = ["None", "FM Radio", "Lamp", "Monitor",
                  "TONE", "Alarm", "Weather"]
LONG_KEY_LIST = ["None", "FM Radio", "Lamp",
                 "Monitor", "TONE", "Alarm", "Weather"]
BUSYLOCK_LIST = ["Off", "On"]
PRESS_NAME = ["stopkey1", "ssidekey1", "ssidekey2",
              "ltopkey2", "lsidekey3", "lsidekey4"]

VFOA_NAME = ["rxfreqa",
             "txfreq",
             "rxtone",
             "txtone",
             "pttid",
             "specialqta",
             "bcl",
             "lowpower",
             "wide",
             "offset"]

VFOB_NAME = ["rxfreqb",
             "txfreq",
             "rxtoneb",
             "txtone",
             "pttid",
             "specialqtb",
             "bclb",
             "lowpowerb",
             "wideb",
             "offsetb"]

# KEY
VOX_GAIN = ["OFF", "1", "2", "3", "4", "5"]
VOX_DELAY = ["1.05s", "2.0s", "3.0s"]
PTTID_VALUES = ["Off", "BOT", "EOT", "BOTH"]
BCLOCK_VALUES = ["Off", "On"]
FREQHOP_VALUES = ["Off", "On"]
SCAN_VALUES = ["Del", "Add"]

# AB CHANNEL
A_OFFSET = ["Off", "-", "+"]
A_TX_POWER = ["Low", "Mid", "High"]
A_BAND = ["Wide", "Narrow"]
A_BUSYLOCK = ["Off", "On"]
A_SPEC_QTDQT = ["Off", "On"]
A_WORKMODE = ["VFO", "VFO+CH", "CH Mode"]

B_OFFSET = ["Off", "-", "+"]
B_TX_POWER = ["Low", "Mid", "High"]
B_BAND = ["Wide", "Narrow"]
B_BUSYLOCK = ["Off", "On"]
B_SPEC_QTDQT = ["Off", "On"]
B_WORKMODE = ["VFO", "VFO+CH", "CH Mode"]

# FM
FM_WORKMODE = ["CH", "VFO"]
FM_CHANNEL = ['%s' % x for x in range(0, 26)]

# DTMF
GROUPCODE = ["", "Off", "*", "#", "A", "B", "C", "D"]

AB_LIST = ["A", "B"]
ALMOD_LIST = ["Site", "Tone", "Code"]
BANDWIDTH_LIST = ["Wide", "Narrow"]
COLOR_LIST = ["Off", "Blue", "Orange", "Purple"]
DTMFSPEED_LIST = ["%s ms" % x for x in range(50, 2010, 10)]
DTMFST_LIST = ["OFF", "DT-ST", "ANI-ST", "DT+ANI"]
MODE_LIST = ["Channel", "Name", "Frequency"]
PONMSG_LIST = ["Full", "Message", "Icon"]
PTTID_LIST = ["Off", "BOT", "EOT", "Both"]
PTTIDCODE_LIST = ["%s" % x for x in range(1, 16)]
RTONE_LIST = ["1000 Hz", "1450 Hz", "1750 Hz", "2100 Hz"]
RESUME_LIST = ["TO", "CO", "SE"]
ROGERRX_LIST = ["Off"] + AB_LIST
RPSTE_LIST = ["OFF"] + ["%s" % x for x in range(1, 11)]
SAVE_LIST = ["Off", "1:1", "1:2", "1:3", "1:4"]
SCODE_LIST = ["%s" % x for x in range(1, 16)]
SHIFTD_LIST = ["Off", "+", "-"]
STEDELAY_LIST = ["OFF"] + ["%s ms" % x for x in range(100, 1100, 100)]
STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 25.0]
STEP_LIST = [str(x) for x in STEPS]
STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 20.0, 25.0, 50.0]
STEP291_LIST = [str(x) for x in STEPS]
TDRAB_LIST = ["Off"] + AB_LIST
TDRCH_LIST = ["CH%s" % x for x in range(1, 129)]
TIMEOUT_LIST = ["%s sec" % x for x in range(15, 615, 15)] + \
    ["Off (if supported by radio)"]
TXPOWER_LIST = ["High", "Low"]
TXPOWER3_LIST = ["High", "Mid", "Low"]
VOICE_LIST = ["Off", "English", "Chinese"]
VOX_LIST = ["OFF"] + ["%s" % x for x in range(1, 11)]
WORKMODE_LIST = ["Frequency", "Channel"]
# mic
MIC_GAIN_LIST = ['%s' % x for x in range(0, 10)]
MIC_GAIN_LIST_H8 = ['%s' % x for x in range(0, 33)]
H8_LIST = ["TD-H8", "TD-H8-HAM", "TD-H8-GMRS"]

GMRS_FREQS = bandplan_na.GMRS_HIRPT

NOAA_FREQS = [162550000, 162400000, 162475000, 162425000, 162450000,
              162500000, 162525000, 161650000, 161775000, 161750000,
              162000000]

HAM_GMRS_NAME = ["NOAA 1", "NOAA 2", "NOAA 3", "NOAA 4", "NOAA 5", "NOAA 6",
                 "NOAA 7", "NOAA 8", "NOAA 9", "NOAA 10", "NOAA 11"]

ALL_MODEL = ["TD-H8", "TD-H8-HAM", "TD-H8-GMRS", "TD-H3",
             "TD-H3-HAM", "TD-H3-GMRS"]

TD_H8 = b"\x50\x56\x4F\x4A\x48\x1C\x14"
TD_H3 = b"\x50\x56\x4F\x4A\x48\x5C\x14"


def _do_status(radio, block):
    status = chirp_common.Status()
    status.msg = "Cloning"
    status.cur = block
    status.max = radio._memsize
    radio.status_fn(status)


def _upper_band_from_data(data):
    return data[0x03:0x04]


def _upper_band_from_image(radio):
    return _upper_band_from_data(radio.get_mmap())


def _firmware_version_from_data(data, version_start, version_stop):
    version_tag = data[version_start:version_stop]
    return version_tag


def _firmware_version_from_image(radio):
    version = _firmware_version_from_data(radio.get_mmap(),
                                          radio._fw_ver_file_start,
                                          radio._fw_ver_file_stop)
    # LOG.debug("_firmware_version_from_image: " + util.hexprint(version))
    return version


def _do_ident(radio, magic, secondack=True):
    serial = radio.pipe
    serial.timeout = 1

    LOG.info("Sending Magic: %s" % util.hexprint(magic))
    serial.write(magic)
    ack = serial.read(1)

    if ack != b"\x06":
        if ack:
            # LOG.debug(repr(ack))
            pass
        raise errors.RadioError("Radio did not respond")

    serial.write(b"\x02")

    response = b""
    for i in range(1, 9):
        byte = serial.read(1)
        response += byte
        if byte == b"\xDD":
            break

    if len(response) in [8, 12]:
        # DEBUG
        LOG.info("Valid response, got this:")
        LOG.info(util.hexprint(response))
        if len(response) == 12:
            ident = response[0] + response[3] + response[5] + response[7:]
        else:
            ident = response
    else:
        # bad response
        msg = "Unexpected response, got this:"
        msg += util.hexprint(response)
        LOG.debug(msg)
        raise errors.RadioError("Unexpected response from radio.")

    if secondack:
        serial.write(b"\x06")
        ack = serial.read(1)
        if ack != b"\x06":
            raise errors.RadioError("Radio refused clone")

    return ident


def response_mode(mode):
    data = mode
    return data


def _read_block(radio, start, size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'R', start, size)
    expectedresponse = b"W" + cmd[1:]

    try:
        serial.write(cmd)
        response = serial.read(5 + size)
        if response[:4] != expectedresponse:
            raise errors.RadioError("Error reading block %04x." % (start))
        block_data = response[4:-1]

    except Exception:
        raise errors.RadioError("Failed to read block at %04x" % start)

    return block_data


def _get_radio_firmware_version(radio):
    if radio.MODEL in ALL_MODEL:
        block = _read_block(radio, 0x1B40, 0x20)
        version = block[0:6]
    return version


IDENT_BLACKLIST = {
    b"\x50\x56\x4F\x4A\x48\x1C\x14": "Radio identifies as TIDRADIO TD-H8",
}


def _ident_radio(radio):
    for magic in radio._idents:
        error = None
        try:
            data = _do_ident(radio, magic)
            return data
        except errors.RadioError as e:
            error = e
            time.sleep(2)

    if error:
        raise error
    raise errors.RadioError("Radio did not respond")


def _do_download(radio):
    data = _ident_radio(radio)
    append_model = False
    # HAM OR GMRS
    # Determine the walkie-talkie mode
    # TDH8 have three mode:ham, gmrs and normal

    LOG.info("Radio mode is " + str(data)[2:8])
    LOG.info("Chirp choose mode is " + str(data)[2:8])
    # The Ham and GMRS modes are subclasses of this model TDH8.
    # We compare the radio identification to the value of that class to
    # make sure the user chose the model that matches
    # the radio we're talking to right now. If they do not match,
    # we refuse to talk to the radio until the user selects the correct model.

    if radio.ident_mode == data:
        LOG.info("Successful match.")
    else:
        msg = ("Model mismatch!")
        raise errors.RadioError(msg)

    # Main block
    LOG.info("Downloading...")

    for i in range(0, radio._memsize, 0x20):
        block = _read_block(radio, i, 0x20)
        data += block
        _do_status(radio, i)
    _do_status(radio, radio._memsize)
    LOG.info("done.")

    if append_model:
        data += radio.MODEL.ljust(8)

    return memmap.MemoryMapBytes(data)


def _exit_write_block(radio):
    serial = radio.pipe
    try:
        serial.write(b"E")

    except Exception:
        raise errors.RadioError("Radio refused to exit programming mode")


def _write_block(radio, addr, data):
    serial = radio.pipe
    cmd = struct.pack(">cHb", b'W', addr, 0x20)
    data = radio.get_mmap()[addr + 8: addr + 40]
    # The checksum needs to be in the last
    check_sum = bytes([sum(data) & 0xFF])
    data += check_sum
    used_data = cmd + data
    serial.write(used_data)

    ack = radio.pipe.read(1)
    if ack != b"\x06":
        raise errors.RadioError("Radio refused to accept block 0x%04x" % addr)


def _do_upload(radio):
    data = _ident_radio(radio)
    radio_version = _get_radio_firmware_version(radio)
    LOG.info("Radio Version is %s" % repr(radio_version))

    if radio.ident_mode == data:
        LOG.info("Successful match.")
    else:
        msg = ("Model mismatch!")
        raise errors.RadioError(msg)

    # Main block
    LOG.debug("Uploading...")

    for start_addr, end_addr in radio._ranges_main:
        for addr in range(start_addr, end_addr, 0x20):
            _write_block(radio, addr, 0x20)
            _do_status(radio, addr)
    _exit_write_block(radio)
    LOG.debug("Upload all done.")


TX_POWER = [chirp_common.PowerLevel("Low",  watts=1.00),
            chirp_common.PowerLevel("Mid",  watts=4.00),
            chirp_common.PowerLevel("High", watts=8.00)]

TDH8_CHARSET = chirp_common.CHARSET_ALPHANUMERIC + \
    "!@#$%^&*()+-=[]:\";'<>?,./"


@directory.register
class TDH8(chirp_common.CloneModeRadio):
    """TIDRADIO TD-H8"""
    VENDOR = "TIDRADIO"
    MODEL = "TD-H8"
    ident_mode = b'P31183\xff\xff'
    BAUD_RATE = 38400
    NEEDS_COMPAT_SERIAL = False
    _memsize = 0x1eef
    _ranges_main = [(0x0000, 0x1eef)]
    _idents = [TD_H8]
    _vhf_range = (136000000, 175000000)
    _uhf_range = (400000000, 521000000)
    _aux_block = True
    _tri_power = True
    _gmrs = False
    _ham = False
    _mem_params = (0x1F2F)

    # offset of fw version in image file
    _fw_ver_file_start = 0x1838
    _fw_ver_file_stop = 0x1846
    _valid_chars = TDH8_CHARSET

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = (dedent("""\
            1. Turn radio off.
            2. Connect cable to mic/spkr connector.
            3. Make sure connector is firmly connected.
            4. Turn radio on (volume may need to be set at 100%).
            5. Ensure that the radio is tuned to channel with no activity.
            6. Click OK to download image from device."""))
        rp.pre_upload = (dedent("""\
            1. Turn radio off.
            2. Connect cable to mic/spkr connector.
            3. Make sure connector is firmly connected.
            4. Turn radio on (volume may need to be set at 100%).
            5. Ensure that the radio is tuned to channel with no activity.
            6. Click OK to upload image to device."""))
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_cross = True
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_tuning_step = False
        rf.has_ctone = True
        rf.can_odd_split = True
        rf.valid_name_length = 8
        rf.valid_characters = self._valid_chars
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_power_levels = TX_POWER
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_tuning_steps = STEPS

        normal_bands = [self._vhf_range, self._uhf_range]

        if self._mmap is None:
            rf.valid_bands = [normal_bands[0], normal_bands[1]]
        else:
            rf.valid_bands = normal_bands
        rf.memory_bounds = (1, 199)
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        try:
            self._mmap = _do_download(self)
            self.process_mmap()
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)

    def sync_out(self):
        try:
            _do_upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    # Encoding processing
    def _decode_tone(self, val):
        if val == 16665 or val == 0:
            return '', None, None
        elif val >= 12000:
            return 'DTCS', val - 12000, 'R'
        elif val >= 8000:
            return 'DTCS', val - 8000, 'N'
        else:
            return 'Tone', val / 10.0, None

    # Decoding processing
    def _encode_tone(self, memval, mode, value, pol):
        if mode == "":
            memval[0].set_raw(0xFF)
            memval[1].set_raw(0xFF)
        elif mode == 'Tone':
            memval.set_value(int(value * 10))

        elif mode == 'DTCS':
            flag = 0x80 if pol == 'N' else 0xC0
            memval.set_value(value)
            memval[1].set_bits(flag)
        else:
            raise Exception("Internal error: invalid mode `%s'" % mode)

    def _get_mem(self, number):
        return self._memobj.memory[number]

    def _get_nam(self, number):
        if self.MODEL in H8_LIST:
            return self._memobj.names[number]
        else:
            return self._memobj.names[number - 1]

    def _get_fm(self, number):
        return self._memobj.fmmode[number]

    def _get_get_scanvfo(self, number):
        return self._memobj.fmvfo[number]

    def _get_scan(self, number):
        return self._memobj.scanadd[number]

    def _get_block_data(self, number):
        return self._memobj.usedflags[number]

    def _get_fmblock_data(self, number):
        return self._memobj.fmusedflags[number]

    def get_memory(self, number):
        _mem = self._get_mem(number)
        _nam = self._get_nam(number)
        mem = chirp_common.Memory()
        mem.number = number

        if _mem.get_raw(asbytes=False)[0] == "\xff":
            mem.empty = True
            return mem

        # receiving frequency
        mem.freq = int(_mem.rxfreq) * 10

        # narrow and wide
        mem.mode = _mem.wide and "NFM" or "FM"

        # power
        try:
            mem.power = TX_POWER[_mem.lowpower]
        except IndexError:
            LOG.error("Radio reported invalid power level %s (in %s)" %
                      (_mem.lowpower, TX_POWER))
            mem.power = TX_POWER[0]

        # Channel name
        for char in _nam.name:
            if "\x00" in str(char) or "\xFF" in str(char):
                char = ""
            mem.name += str(char)

        mem.name = mem.name.rstrip()
        if self.ident_mode != b'P31183\xff\xff' and \
                (mem.number >= 189 and mem.number <= 199):
            mem.name = HAM_GMRS_NAME[mem.number - 200]

        # tmode
        lin2 = int(_mem.rxtone)
        rxtone = self._decode_tone(lin2)

        lin = int(_mem.txtone)
        txtone = self._decode_tone(lin)

        if txtone[0] == "Tone" and not rxtone[0]:
            mem.tmode = "Tone"
        elif txtone[0] == rxtone[0] and txtone[0] == "Tone" \
                and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif txtone[0] == rxtone[0] and txtone[0] == "DTCS" \
                and mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif rxtone[0] or txtone[0]:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txtone[0], rxtone[0])

        chirp_common.split_tone_decode(mem, txtone, rxtone)

        # skip/scanadd
        scan_val_list = []
        for x in range(25):
            a = self._get_scan(x)
            for i in range(0, 8):
                scan_val = (getattr(a, 'scan%i' % (i+1)))
                # print(str(scan_val)[3])
                used_scan_val = str(scan_val)[3]
                scan_val_list.append(used_scan_val)

        if int(scan_val_list[number - 1]) == 0:
            mem.skip = 'S'
        elif int(scan_val_list[number - 1]) == 1:
            mem.skip = ''

        if int(_mem.rxfreq) == int(_mem.txfreq):
            mem.duplex = ""
            mem.offset = 0
        elif int(_mem.txfreq) == 66666665 and int(_mem.rxfreq) != 66666665:
            mem.offset = 0
            mem.duplex = 'off'
        else:
            mem.duplex = int(_mem.rxfreq) > int(_mem.txfreq) and "-" or "+"
            mem.offset = abs(int(_mem.rxfreq) - int(_mem.txfreq)) * 10

        if self._gmrs:
            # mem.duplex = ""
            # mem.offset = 0
            if mem.number >= 1 and mem.number <= 30:
                mem.immutable.append('freq')
                if mem.number >= 8 and mem.number <= 14:
                    mem.mode = 'NFM'
                    mem.power = TX_POWER[0]
                    mem.immutable = ['freq', 'mode', 'power',
                                     'duplex', 'offset']
            elif mem.number >= 31 and mem.number <= 54:
                # mem.immutable = ['duplex', 'offset']
                mem.duplex = '+'
                mem.offset = 5000000
            elif mem.number >= 189 and mem.number <= 199:
                ham_freqs = NOAA_FREQS[mem.number - 189]
                mem.freq = ham_freqs
                mem.immutable = ['name', 'power', 'duplex', 'freq',
                                 'rx_dtcs', 'vfo', 'tmode', 'empty',
                                 'offset', 'rtone', 'ctone', 'dtcs',
                                 'dtcs_polarity', 'cross_mode']
        elif self._ham:
            if mem.number >= 189 and mem.number <= 199:
                ham_freqs = NOAA_FREQS[mem.number - 189]
                mem.freq = ham_freqs
                mem.immutable = ['name', 'power', 'freq', 'rx_dtcs', 'vfo',
                                 'tmode', 'empty', 'offset', 'rtone', 'ctone',
                                 'dtcs', 'dtcs_polarity', 'cross_mode']

        # other function
        # pttid
        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("pttid", "PTT ID",
                          RadioSettingValueList(PTTID_VALUES,
                                                PTTID_VALUES[_mem.pttid]))
        mem.extra.append(rs)

        # Busylock
        rs = RadioSetting("bcl", "Busy Lock",
                          RadioSettingValueList(BCLOCK_VALUES,
                                                BCLOCK_VALUES[_mem.bcl]))
        mem.extra.append(rs)

        rs = RadioSetting(
            "freqhop", "Frequency Hop", RadioSettingValueList(
                FREQHOP_VALUES, FREQHOP_VALUES[_mem.freqhop]))
        mem.extra.append(rs)

        return mem

    def _set_mem(self, number):
        return self._memobj.memory[number]

    def _set_nam(self, number):
        if self.MODEL in H8_LIST:
            return self._memobj.names[number]
        else:
            return self._memobj.names[number - 1]

    def _get_scan_list(self, scan_data):
        # scan_val_list - Get all scans Add data 1-200 digits
        scan_val_list = []
        for x in range(25):
            a = self._get_scan(x)
            for i in range(0, 8):
                scan_val = (getattr(a, 'scan%i' % (i+1)))
                used_scan_val = str(scan_val)[3]
                scan_val_list.append(used_scan_val)

        # used_scan_list - 25 structures, split the scan added
        # data into 25 groups of 8 bits each
        used_scan_list = []
        count_num = 1
        for i in range(0, len(scan_val_list), 8):
            used_scan_list.append(scan_val_list[i:i + 8])
            count_num += 1
        # Determine whether it is a standard number that can be divisible
        # Which group is the scan addition located in the modified channel
        if scan_data % 8 != 0:
            x_list = scan_data / 8
            y_list = scan_data % 8

        else:
            x_list = (scan_data / 8) - 1
            y_list = 8

        return ([x_list, y_list])

    def get_block_xy(self, number):
        if number % 8 != 0:
            x_list = number / 8
            y_list = number % 8

        else:
            x_list = (number / 8) - 1
            y_list = 8

        return ([x_list, y_list])

    def set_memory(self, mem):
        _mem = self._get_mem(mem.number)
        _nam = self._get_nam(mem.number)

        if mem.empty:
            _mem.set_raw("\xff" * 16)
            if self.MODEL in H8_LIST:
                _nam.set_raw("\xff" * 16)
            else:
                _nam.set_raw("\xff" * 8)
            return

        _mem.set_raw("\x00" * 16)

        # When the channel is empty, you need to set "usedflags" to 0,
        # which means it is empty.When the channel has a value,
        # you need to set "usedflags" to 0,
        # indicating that it contains a value.
        # The method "get_block_xy" is due to
        # the particularity of the structure,
        # x represents a certain structure group,
        # and y represents a certain byte of the structure group.
        if not mem.empty:
            block_xy = self.get_block_xy(mem.number)
            _usedflags = self._get_block_data(block_xy[0])
            if block_xy[1] == 1:
                _usedflags.block1 = 1
            elif block_xy[1] == 2:
                _usedflags.block2 = 1
            elif block_xy[1] == 3:
                _usedflags.block3 = 1
            elif block_xy[1] == 4:
                _usedflags.block4 = 1
            elif block_xy[1] == 5:
                _usedflags.block5 = 1
            elif block_xy[1] == 6:
                _usedflags.block6 = 1
            elif block_xy[1] == 7:
                _usedflags.block7 = 1
            elif block_xy[1] == 8:
                _usedflags.block8 = 1
        else:
            block_xy = self.get_block_xy(mem.number)
            _usedflags = self._get_block_data(block_xy[0])
            if block_xy[1] == 1:
                _usedflags.block1 = 0
            elif block_xy[1] == 2:
                _usedflags.block2 = 0
            elif block_xy[1] == 3:
                _usedflags.block3 = 0
            elif block_xy[1] == 4:
                _usedflags.block4 = 0
            elif block_xy[1] == 5:
                _usedflags.block5 = 0
            elif block_xy[1] == 6:
                _usedflags.block6 = 0
            elif block_xy[1] == 7:
                _usedflags.block7 = 0
            elif block_xy[1] == 8:
                _usedflags.block8 = 0

        if mem.duplex == "":
            _mem.rxfreq = _mem.txfreq = mem.freq / 10
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        elif mem.duplex == 'off':
            _mem.txfreq = 166666665
        else:
            _mem.txfreq = mem.freq / 10

        _mem.rxfreq = mem.freq / 10
        _namelength = self.get_features().valid_name_length

        for i in range(_namelength):
            try:
                _nam.name[i] = mem.name[i]
            except IndexError:
                _nam.name[i] = "\xFF"

        txtone, rxtone = chirp_common.split_tone_encode(mem)

        self._encode_tone(_mem.txtone, *txtone)
        self._encode_tone(_mem.rxtone, *rxtone)

        if mem.mode == "FM":
            _mem.wide = 0
        else:
            _mem.wide = 1

        if str(mem.power) == "Low":
            _mem.lowpower = 0
        elif str(mem.power) == "Mid":
            _mem.lowpower = 1
        elif str(mem.power) == "High":
            _mem.lowpower = 10
        else:
            _mem.lowpower = 0

        # Skip/Scanadd Setting
        scanlist = self._get_scan_list(mem.number)
        _scan = self._get_scan(scanlist[0])

        if scanlist[1] == 1:
            _scan.scan1 = mem.skip != "S"
        elif scanlist[1] == 2:
            _scan.scan2 = mem.skip != "S"
        elif scanlist[1] == 3:
            _scan.scan3 = mem.skip != "S"
        elif scanlist[1] == 4:
            _scan.scan4 = mem.skip != "S"
        elif scanlist[1] == 5:
            _scan.scan5 = mem.skip != "S"
        elif scanlist[1] == 6:
            _scan.scan6 = mem.skip != "S"
        elif scanlist[1] == 7:
            _scan.scan7 = mem.skip != "S"
        elif scanlist[1] == 8:
            _scan.scan8 = mem.skip != "S"

        for setting in mem.extra:
            if (self.ident_mode == b'P31185\xff\xff' or
                self.ident_mode == b'P31184\xff\xff') and \
                    mem.number >= 189 and mem.number <= 199:
                if setting.get_name() == 'pttid':
                    setting.value = 'Off'
                    setattr(_mem, setting.get_name(), setting.value)
                elif setting.get_name() == 'bcl':
                    setting.value = 'Off'
                    setattr(_mem, setting.get_name(), setting.value)
                elif setting.get_name() == 'freqhop':
                    setting.value = 'Off'
                    setattr(_mem, setting.get_name(), setting.value)
            else:
                setattr(_mem, setting.get_name(), setting.value)

    def _is_orig(self):
        version_tag = _firmware_version_from_image(self)
        try:
            if b'BFB' in version_tag:
                idx = version_tag.index(b"BFB") + 3
                version = int(version_tag[idx:idx + 3])
                return version < 291
            return False
        except Exception:
            pass
        raise errors.RadioError("Unable to parse version string %s" %
                                version_tag)

    def _my_upper_band(self):
        band_tag = _upper_band_from_image(self)
        return band_tag

    def _get_settings(self):
        _settings = self._memobj.settings
        _press = self._memobj.press
        _aoffset = self._memobj.aoffset
        _boffset = self._memobj.boffset
        _vfoa = self._memobj.vfoa
        _vfob = self._memobj.vfob
        _gcode = self._memobj.groupcode
        _msg = self._memobj.poweron_msg
        basic = RadioSettingGroup("basic", "Basic Settings")
        abblock = RadioSettingGroup("abblock", "A/B Channel")
        fmmode = RadioSettingGroup("fmmode", "FM")
        dtmf = RadioSettingGroup("dtmf", "DTMF")

        # group = RadioSettings(fmmode, dtmf)
        group = RadioSettings(basic, abblock, fmmode, dtmf)

        rs = RadioSetting("squelch", "Squelch Level",
                          RadioSettingValueList(
                              SQUELCH, SQUELCH[_settings.squelch]))
        basic.append(rs)

        rs = RadioSetting("ligcon", "Light Control",
                          RadioSettingValueList(
                              LIGHT_LIST, LIGHT_LIST[_settings.ligcon]))
        basic.append(rs)

        rs = RadioSetting("voiceprompt", "Voice Prompt",
                          RadioSettingValueList(
                              VOICE_PRMPT_LIST, VOICE_PRMPT_LIST[
                                  _settings.voiceprompt]))
        basic.append(rs)

        rs = RadioSetting("keyautolock", "Auto Lock",
                          RadioSettingValueList(
                              AUTOLOCK_LIST, AUTOLOCK_LIST[
                                  _settings.keyautolock]))
        basic.append(rs)

        rs = RadioSetting("mdfa", "MDF-A",
                          RadioSettingValueList(
                              MDFA_LIST, MDFA_LIST[_settings.mdfa]))
        basic.append(rs)

        rs = RadioSetting("mdfb", "MDF-B",
                          RadioSettingValueList(
                              MDFB_LIST, MDFB_LIST[_settings.mdfb]))
        basic.append(rs)

        rs = RadioSetting("sync", "SYNC",
                          RadioSettingValueList(
                              SYNC_LIST, SYNC_LIST[_settings.sync]))
        basic.append(rs)

        rs = RadioSetting("save", "Battery Save",
                          RadioSettingValueList(
                              BTV_SAVER_LIST, BTV_SAVER_LIST[_settings.save]))
        basic.append(rs)

        rs = RadioSetting("dbrx", "Double Rx",
                          RadioSettingValueList(
                              DBRX_LIST, DBRX_LIST[_settings.dbrx]))
        basic.append(rs)

        rs = RadioSetting("astep", "A Step",
                          RadioSettingValueList(
                              ASTEP_LIST, ASTEP_LIST[_settings.astep]))
        basic.append(rs)

        rs = RadioSetting("bstep", "B Step",
                          RadioSettingValueList(
                              BSTEP_LIST, BSTEP_LIST[_settings.bstep]))
        basic.append(rs)

        rs = RadioSetting("scanmode", "Scan Mode",
                          RadioSettingValueList(
                              SCAN_MODE_LIST, SCAN_MODE_LIST[
                                  _settings.scanmode]))
        basic.append(rs)

        rs = RadioSetting("pritx", "Priority TX",
                          RadioSettingValueList(
                              PRIO_LIST, PRIO_LIST[_settings.pritx]))
        basic.append(rs)

        rs = RadioSetting("btnvoice", "Beep",
                          RadioSettingValueBoolean(_settings.btnvoice))
        basic.append(rs)

        rs = RadioSetting("rogerprompt", "Roger",
                          RadioSettingValueBoolean(_settings.rogerprompt))
        basic.append(rs)

        rs = RadioSetting("txled", "Disp Lcd(TX)",

                          RadioSettingValueBoolean(_settings.txled))
        basic.append(rs)

        rs = RadioSetting("rxled", "Disp Lcd(RX)",
                          RadioSettingValueBoolean(_settings.rxled))
        basic.append(rs)

        rs = RadioSetting("onlychmode", "Only CH Mode",
                          RadioSettingValueBoolean(_settings.onlychmode))
        basic.append(rs)
        rs = RadioSetting("ssidekey1", "SHORT_KEY_PF1",
                          RadioSettingValueList(
                              SHORT_KEY_LIST, SHORT_KEY_LIST[
                                  _press.ssidekey1]))
        basic.append(rs)
        rs = RadioSetting("lsidekey3", "LONG_KEY_PF1",
                          RadioSettingValueList(
                              LONG_KEY_LIST,
                              LONG_KEY_LIST[_press.lsidekey3]))
        basic.append(rs)
        if self.MODEL in H8_LIST:
            rs = RadioSetting("stopkey1", "SHORT_KEY_TOP",
                              RadioSettingValueList(SHORT_KEY_LIST,
                                                    SHORT_KEY_LIST[0]))
            basic.append(rs)

            rs = RadioSetting("ltopkey2", "LONG_KEY_TOP",
                              RadioSettingValueList(
                                  LONG_KEY_LIST,
                                  LONG_KEY_LIST[_press.ltopkey2]))
            basic.append(rs)

            rs = RadioSetting("ssidekey2", "SHORT_KEY_PF2",
                              RadioSettingValueList(
                                  SHORT_KEY_LIST,
                                  SHORT_KEY_LIST[_press.ssidekey2]))
            basic.append(rs)

            rs = RadioSetting("lsidekey4", "LONG_KEY_PF2",
                              RadioSettingValueList(
                                LONG_KEY_LIST,
                                LONG_KEY_LIST[_press.lsidekey4]))
            basic.append(rs)

        rs = RadioSetting("voxgain", "VOX Gain",
                          RadioSettingValueList(
                              VOX_GAIN, VOX_GAIN[_settings.voxgain]))
        basic.append(rs)

        rs = RadioSetting("voxdelay", "VOX Delay",
                          RadioSettingValueList(
                              VOX_DELAY, VOX_DELAY[_settings.voxdelay]))
        basic.append(rs)

        rs = RadioSetting("ponmsg", "Power-On Message",
                          RadioSettingValueList(
                              PONMSG_LIST, PONMSG_LIST[_settings.ponmsg]))
        basic.append(rs)

        # mic gain
        if self.MODEL not in H8_LIST:
            _mic = self._memobj.mic
            rs = RadioSetting("micgain", "MIC GAIN",
                              RadioSettingValueList(
                                  MIC_GAIN_LIST, MIC_GAIN_LIST[_mic.micgain]))
            basic.append(rs)

        if self.MODEL not in H8_LIST:
            rs = RadioSetting("kill", "Kill",
                              RadioSettingValueBoolean(_settings.kill))
            basic.append(rs)
            rs = RadioSetting("stun", "Stun",
                              RadioSettingValueBoolean(_settings.stun))
            basic.append(rs)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        rs = RadioSetting("poweron_msg.msg1", "Power-On Message 1",
                          RadioSettingValueString(0, 16, _filter(_msg.msg1)))
        basic.append(rs)
        rs = RadioSetting("poweron_msg.msg2", "Power-On Message 2",
                          RadioSettingValueString(0, 16, _filter(_msg.msg2)))
        basic.append(rs)
        rs = RadioSetting("poweron_msg.msg3", "Power-On Message 3",
                          RadioSettingValueString(0, 16, _filter(_msg.msg3)))
        basic.append(rs)

        # A channel
        a_freq = int(_vfoa.rxfreqa)
        freqa = "%i.%05i" % (a_freq / 100000, a_freq % 100000)
        if freqa == "0.00000":
            val1a = RadioSettingValueString(0, 7, '0.00000')
        else:
            val1a = RadioSettingValueFloat(
                136, 520, float(freqa), 0.00001, 5)
        rs = RadioSetting("rxfreqa", "A Channel - Frequency", val1a)
        abblock.append(rs)

        # Offset
        # If the offset is 12.345
        # Then the data obtained is [0x45, 0x23, 0x01, 0x00]
        a_set_val = _aoffset.ofseta
        a_set_list = len(_aoffset.ofseta) - 1
        real_val = ''
        for i in range(a_set_list, -1, -1):
            real_val += str(a_set_val[i])[2:]
        if real_val == "FFFFFFFF":
            rs = RadioSetting("ofseta", "A Offset Frequency",
                              RadioSettingValueString(0, 7, ""))

        else:
            real_val = int(real_val)
            real_val = "%i.%05i" % (real_val / 100000, real_val % 100000)
            rs = RadioSetting("ofseta", "A Offset Frequency",
                              RadioSettingValueFloat(
                                  0.00000, 59.99750, real_val, 0.00001, 5))
        abblock.append(rs)

        rs = RadioSetting("offset", "A Offset",
                          RadioSettingValueList(
                              A_OFFSET, A_OFFSET[_vfoa.offset]))
        abblock.append(rs)

        rs = RadioSetting("lowpower", "A TX Power",
                          RadioSettingValueList(
                              A_TX_POWER, A_TX_POWER[_vfoa.lowpower]))
        abblock.append(rs)

        rs = RadioSetting("wide", "A Band",
                          RadioSettingValueList(
                              A_BAND, A_BAND[_vfoa.wide]))
        abblock.append(rs)

        rs = RadioSetting("bcl", "A Busy Lock",
                          RadioSettingValueList(
                              A_BUSYLOCK, A_BUSYLOCK[_vfoa.bcl]))
        abblock.append(rs)

        rs = RadioSetting("specialqta", "A Special QT/DQT",
                          RadioSettingValueList(
                              A_SPEC_QTDQT, A_SPEC_QTDQT[_vfoa.specialqta]))
        abblock.append(rs)

        rs = RadioSetting("aworkmode", "A Work Mode",
                          RadioSettingValueList(
                              A_WORKMODE, A_WORKMODE[_settings.aworkmode]))
        abblock.append(rs)

        # B channel
        b_freq = int(str(int(_vfob.rxfreqb)).ljust(8, '0'))
        freqb = "%i.%05i" % (b_freq / 100000, b_freq % 100000)
        if freqb == "0.00000":
            val1a = RadioSettingValueString(0, 7, '0.00000')
        else:
            val1a = RadioSettingValueFloat(
                136, 520, float(freqb), 0.00001, 5)
        rs = RadioSetting("rxfreqb", "B Channel - Frequency", val1a)
        abblock.append(rs)

        # Offset frequency
        # If the offset is 12.345
        # Then the data obtained is [0x45, 0x23, 0x01, 0x00]
        # Need to use the following anonymous function to process data
        b_set_val = _boffset.ofsetb
        b_set_list = len(_boffset.ofsetb) - 1
        real_val = ''
        for i in range(b_set_list, -1, -1):
            real_val += str(b_set_val[i])[2:]
        if real_val == "FFFFFFFF":
            rs = RadioSetting("ofsetb", "B Offset Frequency",
                              RadioSettingValueString(0, 7, " "))
        else:
            real_val = int(real_val)
            real_val = "%i.%05i" % (real_val / 100000, real_val % 100000)
            rs = RadioSetting("ofsetb", "B Offset Frequency",
                              RadioSettingValueFloat(
                                  0.00000, 59.99750, real_val, 0.00001, 5))
        abblock.append(rs)

        rs = RadioSetting("offsetb", "B Offset",
                          RadioSettingValueList(
                              B_OFFSET, B_OFFSET[_vfob.offsetb]))
        abblock.append(rs)

        rs = RadioSetting("lowpowerb", "B TX Power",
                          RadioSettingValueList(
                              B_TX_POWER, B_TX_POWER[_vfob.lowpowerb]))
        abblock.append(rs)

        rs = RadioSetting("wideb", "B Band",
                          RadioSettingValueList(
                              B_BAND, B_BAND[_vfob.wideb]))
        abblock.append(rs)

        rs = RadioSetting("bclb", "B Busy Lock",
                          RadioSettingValueList(
                              B_BUSYLOCK, B_BUSYLOCK[_vfob.bclb]))
        abblock.append(rs)

        rs = RadioSetting("specialqtb", "B Special QT/DQT",
                          RadioSettingValueList(
                              B_SPEC_QTDQT, B_SPEC_QTDQT[_vfob.specialqtb]))
        abblock.append(rs)

        rs = RadioSetting("bworkmode", "B Work Mode",
                          RadioSettingValueList(
                              B_WORKMODE, B_WORKMODE[_settings.bworkmode]))
        abblock.append(rs)

        rs = RadioSetting("fmworkmode", "Work Mode",
                          RadioSettingValueList(
                              FM_WORKMODE, FM_WORKMODE[_settings.fmworkmode]))
        fmmode.append(rs)

        rs = RadioSetting("fmroad", "Channel",
                          RadioSettingValueList(
                              FM_CHANNEL, FM_CHANNEL[_settings.fmroad]))
        fmmode.append(rs)

        rs = RadioSetting("fmrec", "Forbid Receive",
                          RadioSettingValueBoolean(_settings.fmrec))
        fmmode.append(rs)

        # FM
        for i in range(25):
            _fm = self._get_fm(i).fmblock
            _fm_len = len(_fm) - 1
            _fm_temp = ''
            for x in range(_fm_len, -1, -1):
                _fm_temp += str(_fm[x])[2:]
            if _fm_temp == "00000000" or _fm_temp == 'FFFFFFFF':
                rs = RadioSetting('block' + str(i), "Channel" + " " + str(i+1),
                                  RadioSettingValueString(0, 5, ""))
            else:
                _fm_block = int(_fm_temp)
                _fm_block = '%i.%i' % (_fm_block / 10, _fm_block % 10)
                rs = RadioSetting('block' + str(i), "Channel" + " " + str(i+1),
                                  RadioSettingValueString(0, 5, _fm_block))
            fmmode.append(rs)

        _fmv = self._memobj.fmvfo.vfo
        _fmv_len = len(_fmv) - 1
        _fmv_temp = ''
        for x in range(_fmv_len, -1, -1):
            _fmv_temp += str(_fmv[x])[2:]
        _fmv_block = int(_fmv_temp)
        _fmv_block = '%i.%i' % (_fmv_block / 10, _fmv_block % 10)
        rs = RadioSetting(
            "fmvfo", "VFO", RadioSettingValueFloat(
                76.0, 108.0, _fmv_block, 0.1, 1))
        fmmode.append(rs)

        # DTMF
        gcode_val = str(_gcode.gcode)[2:]
        if gcode_val == "FF":
            gcode_val = "Off"
        elif gcode_val == "0F":
            gcode_val = "#"
        elif gcode_val == "0E":
            gcode_val = "*"
        elif gcode_val == '00':
            gcode_val = ""
        else:
            gcode_val = gcode_val[1]
        rs = RadioSetting("gcode", "Group Code",
                          RadioSettingValueList(GROUPCODE,
                                                gcode_val))
        dtmf.append(rs)

        icode_list = self._memobj.icode.idcode
        used_icode = ''
        for i in icode_list:
            if i == 0xFF:
                continue
            used_icode += str(i)[3]
        dtmfcharsani = "0123456789ABCD "
        i_val = RadioSettingValueString(0, 3, used_icode)
        rs = RadioSetting("icode", "ID Code", i_val)
        i_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_1 = self._memobj.group1.group1
        used_group1 = ''
        for i in gcode_list_1:
            if i == 0xFF:
                continue
            used_group1 += str(i)[3]
        group1_val = RadioSettingValueString(0, 7, used_group1)
        rs = RadioSetting("group1", "1", group1_val)
        group1_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_2 = self._memobj.group2.group2
        used_group2 = ''
        for i in gcode_list_2:
            if i == 0xFF:
                continue
            used_group2 += str(i)[3]
        group2_val = RadioSettingValueString(0, 7, used_group2)
        rs = RadioSetting("group2", "2", group2_val)
        group2_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_3 = self._memobj.group3.group3
        used_group3 = ''
        for i in gcode_list_3:
            if i == 0xFF:
                continue
            used_group3 += str(i)[3]
        group3_val = RadioSettingValueString(0, 7, used_group3)
        rs = RadioSetting("group3", "3", group3_val)
        group3_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_4 = self._memobj.group4.group4
        used_group4 = ''
        for i in gcode_list_4:
            if i == 0xFF:
                continue
            used_group4 += str(i)[3]
        group4_val = RadioSettingValueString(0, 7, used_group4)
        rs = RadioSetting("group4", "4", group4_val)
        group4_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_5 = self._memobj.group5.group5
        used_group5 = ''
        for i in gcode_list_5:
            if i == 0xFF:
                continue
            used_group5 += str(i)[3]
        group5_val = RadioSettingValueString(0, 7, used_group5)
        rs = RadioSetting("group5", "5", group5_val)
        group5_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_6 = self._memobj.group6.group6
        used_group6 = ''
        for i in gcode_list_6:
            if i == 0xFF:
                continue
            used_group6 += str(i)[3]
        group6_val = RadioSettingValueString(0, 7, used_group6)
        rs = RadioSetting("group6", "6", group6_val)
        group6_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_7 = self._memobj.group7.group7
        used_group7 = ''
        for i in gcode_list_7:
            if i == 0xFF:
                continue
            used_group7 += str(i)[3]
        group7_val = RadioSettingValueString(0, 7, used_group7)
        rs = RadioSetting("group7", "7", group7_val)
        group7_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        gcode_list_8 = self._memobj.group8.group8
        used_group8 = ''
        for i in gcode_list_8:
            if i == 0xFF:
                continue
            used_group8 += str(i)[3]
        group8_val = RadioSettingValueString(0, 7, used_group8)
        rs = RadioSetting("group8", "8", group7_val)
        group8_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        scode_list = self._memobj.startcode.scode
        used_scode = ''
        for i in scode_list:
            if i == 0xFF:
                continue
            used_scode += str(i)[3]
        scode_val = RadioSettingValueString(0, 7, used_scode)
        rs = RadioSetting("scode", "PTT ID Starting(BOT)", scode_val)
        scode_val.set_charset(dtmfcharsani)
        dtmf.append(rs)

        ecode_list = self._memobj.endcode.ecode
        used_ecode = ''
        for i in ecode_list:
            if i == 0xFF:
                continue
            used_ecode += str(i)[3]
        ecode_val = RadioSettingValueString(0, 7, used_ecode)
        rs = RadioSetting("ecode", "PTT ID Ending(BOT)", ecode_val)
        dtmf.append(rs)
        # H3
        if self.MODEL not in H8_LIST:
            # stuncode
            ecode_list = self._memobj.skcode.stuncode
            used_ecode = ''
            for i in ecode_list:
                if i == 0xFF:
                    continue
                used_ecode += str(i)[3]
            ecode_val = RadioSettingValueString(0, 16, used_ecode)
            rs = RadioSetting("stuncode", "Stun Code", ecode_val)
            dtmf.append(rs)
            # killcode
            ecode_list = self._memobj.skcode.killcode
            used_ecode = ''
            for i in ecode_list:
                if i == 0xFF:
                    continue
                used_ecode += str(i)[3]
            ecode_val = RadioSettingValueString(0, 16, used_ecode)
            rs = RadioSetting("killcode", "Kill Code", ecode_val)
            dtmf.append(rs)

        return group

    def get_settings(self):
        try:
            return self._get_settings()
        except Exception:
            raise InvalidValueError("Setting Failed!")

    def set_settings(self, settings):

        def fm_validate(value):
            if 760 > value or value > 1080:
                msg = ("FM Channel muse be between 76.0-108.0")
                raise InvalidValueError(msg)

        _settings = self._memobj.settings
        _press = self._memobj.press
        _aoffset = self._memobj.aoffset
        _boffset = self._memobj.boffset
        _vfoa = self._memobj.vfoa
        _vfob = self._memobj.vfob
        _fmmode = self._memobj.fmmode

        for element in settings:
            if not isinstance(element, RadioSetting):
                if element.get_name() == "fm_preset":
                    self._set_fm_preset(element)
                else:
                    self.set_settings(element)
                    continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    elif name in PRESS_NAME:
                        obj = _press
                        setting = element.get_name()

                    elif name in VFOA_NAME:
                        obj = _vfoa
                        setting = element.get_name()
                    elif name == "ofseta":
                        obj = _aoffset
                        setting = element.get_name()
                    elif name in VFOB_NAME:
                        obj = _vfob
                        setting = element.get_name()
                    elif name == "ofsetb":
                        obj = _boffset
                        setting = element.get_name()
                    elif "block" in name:
                        obj = _fmmode
                        setting = element.get_name()
                    elif "fmvfo" in name:
                        obj = self._memobj.fmvfo.vfo
                        setting = element.get_name()
                    elif "gcode" in name:
                        obj = self._memobj.groupcode.gcode
                        setting = element.get_name()
                    elif "idcode" in name:
                        obj = self._memobj.icode.idcode
                        setting = element.get_name()
                    elif "scode" in name:
                        obj = self._memobj.startcode.scode
                        setting = element.get_name()
                    elif "ecode" in name:
                        obj = self._memobj.endcode.ecode
                        setting = element.get_name()
                    elif "group1" in name:
                        obj = self._memobj.group1.group1
                        setting = element.get_name()
                    elif "group2" in name:
                        obj = self._memobj.group2.group2
                        setting = element.get_name()
                    elif "group3" in name:
                        obj = self._memobj.group3.group3
                        setting = element.get_name()
                    elif "group4" in name:
                        obj = self._memobj.group4.group4
                        setting = element.get_name()
                    elif "group5" in name:
                        obj = self._memobj.group5.group5
                        setting = element.get_name()
                    elif "group6" in name:
                        obj = self._memobj.group6.group6
                        setting = element.get_name()
                    elif "group7" in name:
                        obj = self._memobj.group7.group7
                        setting = element.get_name()
                    elif "group8" in name:
                        obj = self._memobj.group8.group8
                        setting = element.get_name()
                    elif "micgain" in name:
                        obj = self._memobj.mic.micgain
                        setting = element.get_name()
                    elif "killcode" in name:
                        obj = self._memobj.skcode.killcode
                        setting = element.get_name()
                    elif "stuncode" in name:
                        obj = self._memobj.skcode.stuncode
                        setting = element.get_name()
                    else:
                        obj = _settings
                        setting = element.get_name()
                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()

                    # Channel A
                    elif setting == "rxfreqa" and element.value.get_mutable():
                        val = int(str(element.value).replace(
                            '.', '').ljust(8, '0'))
                        if (val >= 13600000 and val <= 17400000) or \
                                (val >= 40000000 and val <= 52000000):
                            setattr(obj, setting, val)
                        else:
                            msg = (
                                "Frequency must be between "
                                "136.00000-174.00000 or 400.00000-520.00000")
                            raise InvalidValueError(msg)

                    elif setting == "ofseta" and element.value.get_mutable():
                        if '.' in str(element.value):
                            val = str(element.value).replace(' ', '')
                            if len(
                                val[val.index(".") + 1:]
                                ) >= 1 and int(val[val.index(".") + 1:]
                                               ) != 0:
                                val = '00' + val.replace('.', '')
                            else:
                                val = '0' + val.replace('.', '')
                            val = val.ljust(8, '0')
                            lenth_val = 0
                            list_val = []
                        else:
                            val = '0' + str(element.value).replace(' ', '')
                            val = val.ljust(8, '0')
                            lenth_val = 0
                            list_val = []
                        if (int(val) >= 0 and int(val) <= 5999750):
                            if int(val) == 0:
                                _aoffset.ofseta = [0xFF, 0xFF, 0xFF, 0xFF]
                            else:
                                while lenth_val < (len(val)):
                                    list_val.insert(
                                        0, val[lenth_val:lenth_val + 2])
                                    lenth_val += 2
                                for i in range(len(list_val)):
                                    list_val[i] = int(list_val[i], 16)
                                _aoffset.ofseta = list_val
                        else:
                            msg = ("Offset must be between 0.00000-59.99750")
                            raise InvalidValueError(msg)

                    # B channel
                    elif setting == "rxfreqb" and element.value.get_mutable():
                        val = 0
                        val = int(str(element.value).replace(
                            '.', '').ljust(8, '0'))
                        if (val >= 13600000 and val <= 17400000) or \
                                (val >= 40000000 and val <= 52000000):
                            setattr(obj, setting, val)
                        else:
                            msg = (
                                "Frequency must be between "
                                "136.00000-174.00000 or 400.00000-520.00000")
                            raise InvalidValueError(msg)
                        # setattr(obj, setting, val)

                    elif setting == "ofsetb" and element.value.get_mutable():
                        if '.' in str(element.value):
                            val = str(element.value).replace(' ', '')
                            if len(val[val.index(".") + 1:]
                                   ) >= 1 and int(val[val.index(".") + 1:]
                                                  ) != 0:
                                val = '00' + \
                                    str(element.value).replace('.', '')
                            else:
                                val = '0' + str(element.value).replace('.', '')
                            val = val.ljust(8, '0')
                            lenth_val = 0
                            list_val = []
                        else:
                            val = '0' + str(element.value).replace(' ', '')
                            val = val.ljust(8, '0')
                            lenth_val = 0
                            list_val = []
                        if (int(val) >= 0 and int(val) <= 5999750):
                            if int(val) == 0:
                                _boffset.ofsetb = [0xFF, 0xFF, 0xFF, 0xFF]
                            else:
                                while lenth_val < (len(val)):
                                    list_val.insert(
                                        0, val[lenth_val:lenth_val + 2])
                                    lenth_val += 2
                                for i in range(len(list_val)):
                                    list_val[i] = int(list_val[i], 16)
                                _boffset.ofsetb = list_val
                        else:
                            msg = ("Offset must be between 0.00000-59.99750")
                            raise InvalidValueError(msg)

                    # FM
                    elif "block" in name:

                        val = str(element.value).replace('.', '').zfill(8)
                        num = int(element.get_name().replace('block', '')) + 1
                        lenth_val = 0
                        list_val = []
                        if str(element.value)[0] == ' ':
                            list_val = [0, 0, 0, 0]
                        else:
                            val = val.replace(' ', '').zfill(8)
                            fm_validate(int(val))
                            while lenth_val < (len(val)):
                                list_val.insert(
                                    0, int(val[lenth_val:lenth_val + 2], 16))
                                lenth_val += 2
                        self._memobj.fmmode[num - 1].fmblock = list_val

                        if val != '00000   ':
                            fmblock_xy = self.get_block_xy(num)
                            _fmusedflags = self._get_fmblock_data(
                                fmblock_xy[0])
                            if fmblock_xy[1] == 1:
                                _fmusedflags.block1 = 1
                            elif fmblock_xy[1] == 2:
                                _fmusedflags.block2 = 1
                            elif fmblock_xy[1] == 3:
                                _fmusedflags.block3 = 1
                            elif fmblock_xy[1] == 4:
                                _fmusedflags.block4 = 1
                            elif fmblock_xy[1] == 5:
                                _fmusedflags.block5 = 1
                            elif fmblock_xy[1] == 6:
                                _fmusedflags.block6 = 1
                            elif fmblock_xy[1] == 7:
                                _fmusedflags.block7 = 1
                            elif fmblock_xy[1] == 8:
                                _fmusedflags.block8 = 1
                        else:
                            fmblock_xy = self.get_block_xy(num)
                            _fmusedflags = self._get_fmblock_data(
                                fmblock_xy[0])
                            if fmblock_xy[1] == 1:
                                _fmusedflags.block1 = 0
                            elif fmblock_xy[1] == 2:
                                _fmusedflags.block2 = 0
                            elif fmblock_xy[1] == 3:
                                _fmusedflags.block3 = 0
                            elif fmblock_xy[1] == 4:
                                _fmusedflags.block4 = 0
                            elif fmblock_xy[1] == 5:
                                _fmusedflags.block5 = 0
                            elif fmblock_xy[1] == 6:
                                _fmusedflags.block6 = 0
                            elif fmblock_xy[1] == 7:
                                _fmusedflags.block7 = 0
                            elif fmblock_xy[1] == 8:
                                _fmusedflags.block8 = 0

                    elif setting == 'fmvfo' and element.value.get_mutable():
                        val = str(element.value).replace('.', '').zfill(8)
                        lenth_val = 0
                        list_val = []
                        if " " in val:
                            list_val = [0, 0, 0, 0]
                        else:
                            while lenth_val < (len(val)):
                                list_val.insert(
                                    0, int(val[lenth_val:lenth_val + 2], 16))
                                lenth_val += 2
                        self._memobj.fmvfo.vfo = list_val

                    elif setting == 'gcode' and element.value.get_mutable():
                        val = str(element.value)
                        if val == 'Off':
                            gcode_used = 0xFF
                        elif val == 'A':
                            gcode_used = 0x0A
                        elif val == 'B':
                            gcode_used = 0x0B
                        elif val == 'C':
                            gcode_used = 0x0C
                        elif val == 'D':
                            gcode_used = 0x0D
                        elif val == '#':
                            gcode_used = 0x0F
                        elif val == '*':
                            gcode_used = 0x0E
                        elif val == '':
                            gcode_used = 0x00
                        self._memobj.groupcode.gcode = gcode_used

                    elif setting == 'icode' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.icode.idcode = list_val

                    elif setting == 'scode' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.startcode.scode = list_val

                    elif setting == 'ecode' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.endcode.ecode = list_val

                    elif setting == 'group1' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group1.group1 = list_val

                    elif setting == 'group2' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group2.group2 = list_val

                    elif setting == 'group3' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group3.group3 = list_val

                    elif setting == 'group4' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group4.group4 = list_val

                    elif setting == 'group5' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group5.group5 = list_val

                    elif setting == 'group6' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group6.group6 = list_val

                    elif setting == 'group7' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group7.group7 = list_val

                    elif setting == 'group8' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.group8.group8 = list_val
                    elif setting == 'micgain':
                        self._memobj.mic.micgain = (
                            str(element.value))
                    elif setting == 'stuncode' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.skcode.stuncode = list_val
                    elif setting == 'killcode' and element.value.get_mutable():
                        val = str(element.value)
                        list_val = []
                        lenth_val = 0
                        while lenth_val < (len(val)):
                            if val[lenth_val] != ' ':
                                list_val.append(int(val[lenth_val], 16))
                                lenth_val += 1
                            else:
                                list_val.append(0xFF)
                                lenth_val += 1
                        self._memobj.skcode.killcode = list_val
                    elif element.value.get_mutable():
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    def _set_fm_preset(self, settings):
        for element in settings:
            try:
                val = element.value
                if self._memobj.fm_presets <= 108.0 * 10 - 650:
                    value = int(val.get_value() * 10 - 650)
                else:
                    value = int(val.get_value() * 10)
                LOG.debug("Setting fm_presets = %s" % (value))
                self._memobj.fm_presets = value
            except Exception:
                LOG.debug(element.get_name())
                raise


@directory.register
class TDH8_HAM(TDH8):
    VENDOR = "TIDRADIO"
    MODEL = "TD-H8-HAM"
    ident_mode = b'P31185\xff\xff'
    _ham = True
    _vhf_range = (144000000, 149000000)
    _uhf_range = (420000000, 451000000)


@directory.register
class TDH8_GMRS(TDH8):
    VENDOR = "TIDRADIO"
    MODEL = "TD-H8-GMRS"
    ident_mode = b'P31184\xff\xff'
    _gmrs = True
    _vhf_range = (136000000, 175000000)
    _uhf_range = (400000000, 521000000)

    def validate_memory(self, mem):
        msgs = super().validate_memory(mem)
        if 31 <= mem.number <= 54 and mem.freq not in GMRS_FREQS:
            msgs.append(chirp_common.ValidationError(
                "The frequency in channels 31-54 must be between"
                "462.55000-462.72500 in 0.025 increments."))
        return msgs


@directory.register
class UV68(TDH8):
    VENDOR = "TID"
    MODEL = "TD-UV68"


@directory.register
class TDH3(TDH8):
    VENDOR = "TIDRADIO"
    MODEL = "TD-H3"
    _memsize = 0x1fef
    _ranges_main = [(0x0000, 0x1fef)]
    _idents = [TD_H3]
    _vhf_range = (136000000, 175000000)
    _uhf_range = (400000000, 521000000)
    _aux_block = True
    _tri_power = True
    _gmrs = False
    _ham = False
    _mem_params = (0x1F2F)

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT_H3, self._mmap)


@directory.register
class TDH3_HAM(TDH3):
    VENDOR = "TIDRADIO"
    MODEL = "TD-H3-HAM"
    ident_mode = b'P31185\xff\xff'
    _ham = True
    _vhf_range = (144000000, 149000000)
    _uhf_range = (420000000, 451000000)


@directory.register
class TDH3_GMRS(TDH3):
    VENDOR = "TIDRADIO"
    MODEL = "TD-H3-GMRS"
    ident_mode = b'P31184\xff\xff'
    _gmrs = True
    _vhf_range = (136000000, 175000000)
    _uhf_range = (400000000, 521000000)

    def validate_memory(self, mem):
        msgs = super().validate_memory(mem)
        if 31 <= mem.number <= 54 and mem.freq not in GMRS_FREQS:
            msgs.append(chirp_common.ValidationError(
                "The frequency in channels 31-54 must be between"
                "462.55000-462.72500 in 0.025 increments."))
        return msgs



# Copyright 2011 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""TYT uv3r radio management module"""

import logging
from chirp import chirp_common, bitwise, errors, directory
from chirp.drivers.wouxun import do_download, do_upload

LOG = logging.getLogger(__name__)


def tyt_uv3r_prep(radio):
    try:
        radio.pipe.write(b"PROGRAMa")
        ack = radio.pipe.read(1)
        if ack != b"\x06":
            raise errors.RadioError("Radio did not ACK first command")
    except:
        raise errors.RadioError("Unable to communicate with the radio")


def tyt_uv3r_download(radio):
    tyt_uv3r_prep(radio)
    return do_download(radio, 0x0000, 0x0910, 0x0010)


def tyt_uv3r_upload(radio):
    tyt_uv3r_prep(radio)
    return do_upload(radio, 0x0000, 0x0910, 0x0010)


mem_format = """
struct memory {
  ul24 duplex:2,
       bit:1,
       iswide:1,
       bits:2,
       is625:1,
       freq:17;
  ul16 offset;
  ul16 rx_tone;
  ul16 tx_tone;
  u8 unknown;
  u8 name[6];
};

#seekto 0x0010;
struct memory memory[128];

#seekto 0x0870;
u8 emptyflags[16];
u8 skipflags[16];
"""

THUV3R_DUPLEX = ["", "+", "-"]
THUV3R_CHARSET = "".join([chr(ord("0") + x) for x in range(0, 10)] +
                         [" -*+"] +
                         [chr(ord("A") + x) for x in range(0, 26)] +
                         ["_/"])


@directory.register
class TYTUV3RRadio(chirp_common.CloneModeRadio):
    VENDOR = "TYT"
    MODEL = "TH-UV3R"
    BAUD_RATE = 2400
    _memsize = 2320
    NEEDS_COMPAT_SERIAL = False

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.has_cross = True
        rf.memory_bounds = (1, 128)
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone",
                                "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS"]
        rf.valid_skips = []
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_name_length = 6
        rf.valid_characters = THUV3R_CHARSET
        rf.valid_bands = [(136000000, 520000000)]
        rf.valid_tuning_steps = [5.0, 6.25, 10.0, 12.5, 25.0, 37.50,
                                 50.0, 100.0]
        rf.valid_skips = ["", "S"]
        return rf

    def sync_in(self):
        self.pipe.timeout = 2
        self._mmap = tyt_uv3r_download(self)
        self.process_mmap()

    def sync_out(self):
        tyt_uv3r_upload(self)

    def process_mmap(self):
        self._memobj = bitwise.parse(mem_format, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _decode_tone_value(self, value):
        if value == 0xFFFF:
            return "", "N", 0
        elif value & 0x8000:
            # FIXME: rev pol
            pol = value & 0x4000 and "R" or "N"
            return "DTCS", pol, int("%x" % (value & 0x0FFF))
        else:
            return "Tone", "N", int("%x" % value) / 10.0

    def _decode_tone(self, mem, _mem):
        tx_mode, tpol, tx_tone = self._decode_tone_value(_mem.tx_tone)
        rx_mode, rpol, rx_tone = self._decode_tone_value(_mem.rx_tone)

        if rx_mode == tx_mode == "":
            return

        mem.dtcs_polarity = "%s%s" % (tpol, rpol)

        if rx_mode == tx_mode == "DTCS":
            # Break this for now until we can support this in chirp
            tx_tone = rx_tone

        if rx_mode in ["", tx_mode] and rx_tone in [0, tx_tone]:
            mem.tmode = rx_mode == "Tone" and "TSQL" or tx_mode
            if mem.tmode == "DTCS":
                mem.dtcs = tx_tone
            elif mem.tmode == "TSQL":
                mem.ctone = tx_tone
            else:
                mem.rtone = tx_tone
            return

        mem.cross_mode = "%s->%s" % (tx_mode, rx_mode)
        mem.tmode = "Cross"
        if tx_mode == "Tone":
            mem.rtone = tx_tone
        elif tx_mode == "DTCS":
            mem.dtcs = tx_tone
        if rx_mode == "Tone":
            mem.ctone = rx_tone
        elif rx_mode == "DTCS":
            mem.dtcs = rx_tone  # No support for different codes yet

    def _encode_tone(self, mem, _mem):
        if mem.tmode == "":
            _mem.tx_tone = _mem.rx_tone = 0xFFFF
            return

        def _tone(val):
            return int("%i" % (val * 10), 16)

        def _dcs(val, pol):
            polmask = pol == "R" and 0xC000 or 0x8000
            return int("%i" % (val), 16) | polmask

        rx_tone = tx_tone = 0xFFFF

        if mem.tmode == "Tone":
            rx_mode = ""
            tx_mode = "Tone"
            tx_tone = _tone(mem.rtone)
        elif mem.tmode == "TSQL":
            rx_mode = tx_mode = "Tone"
            rx_tone = tx_tone = _tone(mem.ctone)
        elif mem.tmode == "DTCS":
            rx_mode = tx_mode = "DTCS"
            tx_tone = _dcs(mem.dtcs, mem.dtcs_polarity[0])
            rx_tone = _dcs(mem.dtcs, mem.dtcs_polarity[1])
        elif mem.tmode == "Cross":
            tx_mode, rx_mode = mem.cross_mode.split("->", 1)
            if tx_mode == "DTCS":
                tx_tone = _dcs(mem.dtcs, mem.dtcs_polarity[0])
            elif tx_mode == "Tone":
                tx_tone = _tone(mem.rtone)
            if rx_mode == "DTCS":
                rx_tone = _dcs(mem.dtcs, mem.dtcs_polarity[1])
            elif rx_mode == "Tone":
                rx_tone = _tone(mem.ctone)

        _mem.rx_tone = rx_tone
        _mem.tx_tone = tx_tone

    def get_memory(self, number):
        _mem = self._memobj.memory[number - 1]
        mem = chirp_common.Memory()
        mem.number = number

        bit = 1 << ((number - 1) % 8)
        byte = (number - 1) / 8

        if self._memobj.emptyflags[byte] & bit:
            mem.empty = True
            return mem

        mult = _mem.is625 and 6250 or 5000
        mem.freq = _mem.freq * mult
        mem.offset = _mem.offset * 5000
        mem.duplex = THUV3R_DUPLEX[_mem.duplex]
        mem.mode = _mem.iswide and "FM" or "NFM"
        self._decode_tone(mem, _mem)
        mem.skip = (self._memobj.skipflags[byte] & bit) and "S" or ""

        for char in _mem.name:
            try:
                c = THUV3R_CHARSET[char]
            except:
                c = ""
            mem.name += c
        mem.name = mem.name.rstrip()

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]
        bit = 1 << ((mem.number - 1) % 8)
        byte = (mem.number - 1) // 8

        if mem.empty:
            self._memobj.emptyflags[byte] |= bit
            _mem.set_raw("\xFF" * 16)
            return

        self._memobj.emptyflags[byte] &= ~bit

        if chirp_common.is_fractional_step(mem.freq):
            mult = 6250
            _mem.is625 = True
        else:
            mult = 5000
            _mem.is625 = False
        _mem.freq = mem.freq / mult
        _mem.offset = mem.offset / 5000
        _mem.duplex = THUV3R_DUPLEX.index(mem.duplex)
        _mem.iswide = mem.mode == "FM"
        self._encode_tone(mem, _mem)

        if mem.skip:
            self._memobj.skipflags[byte] |= bit
        else:
            self._memobj.skipflags[byte] &= ~bit

        name = []
        for char in mem.name.ljust(6):
            try:
                c = THUV3R_CHARSET.index(char)
            except:
                c = THUV3R_CHARSET.index(" ")
            name.append(c)
        _mem.name = name
        LOG.debug(repr(_mem))

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == 2320



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import time
import logging

from chirp import memmap, chirp_common, errors

LOG = logging.getLogger(__name__)

POS_MODE = 5
POS_DUP = 6
POS_TMODE = 6
POS_RTONE = 7
POS_CTONE = 8
POS_DTCS = 9
POS_OFFSET = 10

MEM_LOC_BASE = 0x1700
MEM_LOC_SIZE = 16
MEM_TAG_BASE = 0x5800
MEM_FLG_BASE = 0x0E00

V71_SPECIAL = {}

for i in range(0, 10):
    V71_SPECIAL["L%i" % i] = 1000 + (i * 2)
    V71_SPECIAL["U%i" % i] = 1000 + (i * 2) + 1
for i in range(0, 10):
    V71_SPECIAL["WX%i" % (i + 1)] = 1020 + i
V71_SPECIAL["C VHF"] = 1030
V71_SPECIAL["C UHF"] = 1031

V71_SPECIAL_REV = {}
for k, v in V71_SPECIAL.items():
    V71_SPECIAL_REV[v] = k


def command(s, cmd, timeout=0.5):
    start = time.time()

    data = ""
    LOG.debug("PC->V71: %s" % cmd)
    s.write(cmd + "\r")
    while not data.endswith("\r") and (time.time() - start) < timeout:
        data += s.read(1)
    LOG.debug("V71->PC: %s" % data.strip())
    return data.strip()


def get_id(s):
    r = command(s, "ID")
    if r.startswith("ID "):
        return r.split(" ")[1]
    else:
        raise errors.RadioError("No response to ID command")


EXCH_R = "R\x00\x00\x00"
EXCH_W = "W\x00\x00\x00"


def read_block(s, block, count=256):
    s.write(struct.pack("<cHB", "R", block, 0))
    r = s.read(4)
    if len(r) != 4:
        raise Exception("Did not receive block response")

    cmd, _block, zero = struct.unpack("<cHB", r)
    if cmd != "W" or _block != block:
        raise Exception("Invalid response: %s %i" % (cmd, _block))

    data = ""
    while len(data) < count:
        data += s.read(count - len(data))

    s.write(chr(0x06))
    if s.read(1) != chr(0x06):
        raise Exception("Did not receive post-block ACK!")

    return data


def write_block(s, block, map):
    s.write(struct.pack("<cHB", "W", block, 0))
    base = block * 256
    s.write(map[base:base+256])

    ack = s.read(1)

    return ack == chr(0x06)


def download(radio):
    if command(radio.pipe, "0M PROGRAM") != "0M":
        raise errors.RadioError("No response from radio")

    data = ""
    for i in range(0, 0x7F):
        data += read_block(radio.pipe, i)
        if radio.status_fn:
            s = chirp_common.Status()
            s.msg = "Cloning from radio"
            s.max = 256 * 0x7E
            s.cur = len(data)
            radio.status_fn(s)

    radio.pipe.write("E")

    return memmap.MemoryMap(data)


def upload(radio):
    if command(radio.pipe, "0M PROGRAM") != "0M":
        raise errors.RadioError("No response from radio")

    for i in range(0, 0x7F):
        r = write_block(radio.pipe, i, radio._mmap)
        if not r:
            raise errors.RadioError("Radio NAK'd block %i" % i)
        if radio.status_fn:
            s = chirp_common.Status()
            s.msg = "Cloning to radio"
            s.max = 256 * 0x7E
            s.cur = 256 * i
            radio.status_fn(s)

    radio.pipe.write("E")


def get_mem_offset(number):
    return MEM_LOC_BASE + (MEM_LOC_SIZE * number)


def get_raw_mem(map, number):
    base = get_mem_offset(number)
    # LOG.debug("Offset for %i is %04x" % (number, base))
    return map[base:base+MEM_LOC_SIZE]


def get_used(map, number):
    pos = MEM_FLG_BASE + (number * 2)
    flag = ord(map[pos])
    LOG.debug("Flag byte is %02x" % flag)
    return not (flag & 0x80)


def set_used(map, number, freq):
    pos = MEM_FLG_BASE + (number * 2)
    if freq == 0:
        # Erase
        map[pos] = "\xff\xff"
    elif int(freq / 100) == 1:
        map[pos] = "\x05\x00"
    elif int(freq / 100) == 4:
        map[pos] = "\x08\x00"


def get_skip(map, number):
    pos = MEM_FLG_BASE + (number * 2)
    flag = ord(map[pos+1])
    if flag & 0x01:
        return "S"
    else:
        return ""


def set_skip(map, number, skip):
    pos = MEM_FLG_BASE + (number * 2)
    flag = ord(map[pos+1])
    if skip:
        flag |= 0x01
    else:
        flag &= ~0x01
    map[pos+1] = flag


def get_freq(mmap):
    freq, = struct.unpack("<I", mmap[0:4])
    return freq / 1000000.0


def set_freq(mmap, freq):
    mmap[0] = struct.pack("<I", int(freq * 1000000))


def get_name(map, number):
    base = MEM_TAG_BASE + (8 * number)
    return map[base:base+6].replace("\xff", "")


def set_name(mmap, number, name):
    base = MEM_TAG_BASE + (8 * number)
    mmap[base] = name.ljust(6)[:6].upper()


def get_tmode(mmap):
    val = ord(mmap[POS_TMODE]) & 0x70

    tmodemap = {
        0x00: "",
        0x40: "Tone",
        0x20: "TSQL",
        0x10: "DTCS",
        }

    return tmodemap[val]


def set_tmode(mmap, tmode):
    val = ord(mmap[POS_TMODE]) & 0x8F

    tmodemap = {
        "":     0x00,
        "Tone": 0x40,
        "TSQL": 0x20,
        "DTCS": 0x10,
        }

    mmap[POS_TMODE] = val | tmodemap[tmode]


def get_tone(mmap, offset):
    val = ord(mmap[offset])

    return chirp_common.TONES[val]


def set_tone(mmap, tone, offset):
    LOG.debug(tone)
    mmap[offset] = chirp_common.TONES.index(tone)


def get_dtcs(mmap):
    val = ord(mmap[POS_DTCS])

    return chirp_common.DTCS_CODES[val]


def set_dtcs(mmap, dtcs):
    mmap[POS_DTCS] = chirp_common.DTCS_CODES.index(dtcs)


def get_duplex(mmap):
    val = ord(mmap[POS_DUP]) & 0x03

    dupmap = {
        0x00: "",
        0x01: "+",
        0x02: "-",
        }

    return dupmap[val]


def set_duplex(mmap, duplex):
    val = ord(mmap[POS_DUP]) & 0xFC

    dupmap = {
        "":  0x00,
        "+": 0x01,
        "-": 0x02,
        }

    mmap[POS_DUP] = val | dupmap[duplex]


def get_offset(mmap):
    val, = struct.unpack("<I", mmap[POS_OFFSET:POS_OFFSET+4])
    return val / 1000000.0


def set_offset(mmap, offset):
    mmap[POS_OFFSET] = struct.pack("<I", int(offset * 1000000))


def get_mode(mmap):
    val = ord(mmap[POS_MODE]) & 0x03
    modemap = {
        0x00: "FM",
        0x01: "NFM",
        0x02: "AM",
        }

    return modemap[val]


def set_mode(mmap, mode):
    val = ord(mmap[POS_MODE]) & 0xFC
    modemap = {
        "FM":  0x00,
        "NFM": 0x01,
        "AM":  0x02,
        }

    mmap[POS_MODE] = val | modemap[mode]


def get_memory(map, number):
    if number < 0 or number > (max(V71_SPECIAL.values()) + 1):
        raise errors.InvalidMemoryLocation("Number must be between 0 and 999")

    mem = chirp_common.Memory()
    mem.number = number

    if number > 999:
        mem.extd_number = V71_SPECIAL_REV[number]
    if not get_used(map, number):
        mem.empty = True
        return mem

    mmap = get_raw_mem(map, number)

    mem.freq = get_freq(mmap)
    mem.name = get_name(map, number)
    mem.tmode = get_tmode(mmap)
    mem.rtone = get_tone(mmap, POS_RTONE)
    mem.ctone = get_tone(mmap, POS_CTONE)
    mem.dtcs = get_dtcs(mmap)
    mem.duplex = get_duplex(mmap)
    mem.offset = get_offset(mmap)
    mem.mode = get_mode(mmap)

    if number < 999:
        mem.skip = get_skip(map, number)

    if number > 999:
        mem.immutable = ["number", "bank", "extd_number", "name"]
    if number > 1020 and number < 1030:
        mem.immutable += ["freq"]  # FIXME: ALL

    return mem


def initialize(mmap):
    mmap[0] = \
        "\x80\xc8\xb3\x08\x00\x01\x00\x08" + \
        "\x08\x00\xc0\x27\x09\x00\x00\xff"


def set_memory(map, mem):
    if mem.number < 0 or mem.number > (max(V71_SPECIAL.values()) + 1):
        raise errors.InvalidMemoryLocation("Number must be between 0 and 999")

    mmap = memmap.MemoryMap(get_raw_mem(map, mem.number))

    if not get_used(map, mem.number):
        initialize(mmap)

    set_freq(mmap, mem.freq)
    if mem.number < 999:
        set_name(map, mem.number, mem.name)
    set_tmode(mmap, mem.tmode)
    set_tone(mmap, mem.rtone, POS_RTONE)
    set_tone(mmap, mem.ctone, POS_CTONE)
    set_dtcs(mmap, mem.dtcs)
    set_duplex(mmap, mem.duplex)
    set_offset(mmap, mem.offset)
    set_mode(mmap, mem.mode)

    base = get_mem_offset(mem.number)
    map[base] = mmap.get_packed()

    set_used(map, mem.number, mem.freq)
    if mem.number < 999:
        set_skip(map, mem.number, mem.skip)

    return map


if __name__ == "__main__":
    import sys
    import serial
    s = serial.Serial(port=sys.argv[1], baudrate=9600, dsrdtr=True,
                      timeout=0.25)
    # s.write("\r\r")
    # print get_id(s)
    data = download(s)
    open(sys.argv[2], "wb").write(data)



# Copyright 2012 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import yaesu_clone
from chirp import chirp_common, bitwise

# This driver is unfinished and therefore does not register itself with Chirp.
#
# Downloads should work consistently but an upload has not been attempted.
#
# There is a stray byte at 0xC in the radio download that is not present in the
# save file from the CE-21 software. This puts the first few bytes of channel 1
# in the wrong location. A quick hack to fix the subsequent channels was to
# insert the #seekto dynamically, but this does nothing to fix reading of
# channel 1 (memory[0]).
MEM_FORMAT = """
u8 unknown1[6];
u8 prioritych;

#seekto %d;
struct {
  u8 empty:1,
     txinhibit:1,
     tot:1,
     low_power:1,
     bclo:1,
     btlo:1,
     skip:1,
     pwrsave:1;
  u8 unknown2:5,
     narrow:1,
     unknown2b:2;
  u24 name;
  u8 ctone;
  u8 rtone;
  u8 unknown3;
  bbcd freq_rx[3];
  bbcd freq_tx[3];
} memory[32];

char imgname[10];
"""

STEPS = [5.0, 6.25]
CHARSET = "".join([chr(x) for x in range(ord("0"), ord("9")+1)] +
                  [chr(x) for x in range(ord("A"), ord("Z")+1)]) + \
                      r"<=>*+-\/_ "
TONES = list(chirp_common.TONES)
TONES.remove(165.5)
TONES.remove(171.3)
TONES.remove(177.3)
POWER_LEVELS = [chirp_common.PowerLevel("Hi", watts=5.00),
                chirp_common.PowerLevel("Low", watts=1.0)]


# @directory.register
class VX510Radio(yaesu_clone.YaesuCloneModeRadio):
    """Vertex VX-510V"""
    BAUD_RATE = 9600
    VENDOR = "Vertex Standard"
    MODEL = "VX-510V"

    _model = ""
    _memsize = 470
    _block_lengths = [10, 460]
    _block_size = 8

    def _checksums(self):
        return []
        # These checksums don't pass, so the alg might be different than Yaesu.
        # return [yaesu_clone.YaesuChecksum(0, self._memsize - 2)]
        # return [yaesu_clone.YaesuChecksum(0, 10),
        #         yaesu_clone.YaesuChecksum(12, self._memsize - 1)]

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.can_odd_split = True
        rf.has_bank = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = False
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.memory_bounds = (1, 32)
        rf.valid_bands = [(13600000, 174000000)]
        rf.valid_skips = ["", "S"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_name_length = 4
        rf.valid_characters = CHARSET
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT % 0xA, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number-1])

    def get_memory(self, number):
        mem = chirp_common.Memory()
        mem.number = number

        _mem = self._memobj.memory[number-1]

        mem.empty = _mem.empty
        mem.freq = chirp_common.fix_rounded_step(int(_mem.freq_rx) * 1000)

        for i in range(0, 4):
            index = (_mem.name >> (i*6)) & 0x3F
            mem.name += CHARSET[index]

        freq_tx = chirp_common.fix_rounded_step(int(_mem.freq_tx) * 1000)
        if _mem.txinhibit:
            mem.duplex = "off"
        elif mem.freq == freq_tx:
            mem.duplex = ""
            mem.offset = 0
        elif 144000000 <= mem.freq < 148000000:
            mem.duplex = "+" if freq_tx > mem.freq else "-"
            mem.offset = abs(mem.freq - freq_tx)
        else:
            mem.duplex = "split"
            mem.offset = freq_tx

        mem.mode = _mem.narrow and "NFM" or "FM"
        mem.power = POWER_LEVELS[_mem.low_power]

        rtone = int(_mem.rtone)
        ctone = int(_mem.ctone)
        tmode_tx = tmode_rx = ""

        if rtone & 0x80:
            tmode_tx = "DTCS"
            mem.dtcs = chirp_common.DTCS_CODES[int(rtone) - 0x80]
        elif rtone:
            tmode_tx = "Tone"
            mem.rtone = TONES[rtone - 1]
            if not ctone:
                # not used, but this is a better default than 88.5
                mem.ctone = TONES[rtone - 1]

        if ctone & 0x80:
            tmode_rx = "DTCS"
            mem.rx_dtcs = chirp_common.DTCS_CODES[int(ctone) - 0x80]
        elif ctone:
            tmode_rx = "Tone"
            mem.ctone = TONES[ctone - 1]

        if tmode_tx == "Tone" and not tmode_rx:
            mem.tmode = "Tone"
        elif tmode_tx == tmode_rx and tmode_tx == "Tone" and \
                mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif tmode_tx == tmode_rx and tmode_tx == "DTCS" and \
                mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif tmode_rx or tmode_tx:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (tmode_tx, tmode_rx)

        mem.skip = _mem.skip and "S" or ""

        return mem

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == cls._memsize


# @directory.register
class VX510File(VX510Radio, chirp_common.FileBackedRadio):
    """Vertex CE-21 File"""
    VENDOR = "Vertex Standard"
    MODEL = "CE-21 File"

    _model = ""
    _memsize = 664

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0, self._memsize - 1)]

    def process_mmap(self):
        # CE-21 file is missing the 0xC byte, probably a checksum.
        # It's not a YaesuChecksum.
        self._memobj = bitwise.parse(MEM_FORMAT % 0x9, self._mmap)

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == cls._memsize
